from guppylang import guppy
from guppylang.std.builtins import comptime, nat


@guppy
def foo(n: nat @comptime) -> nat:
    return n


@guppy
def main(b: bool) -> nat:
    return foo(nat(1) if b else nat(2))


main.compile()



"""Dummy module used in `test_imports.py`"""

import hugr.tys as ht
from guppylang import guppy
from guppylang_internals.decorator import custom_type


@guppy
def f(x: int) -> int:
    return x + 1


@guppy.declare
def g() -> int: ...


@custom_type(ht.Bool)
class MyType:
    @guppy.declare
    def __neg__(self: "MyType") -> "MyType": ...



import importlib.util
import inspect
import pathlib
import re
import sys

import pytest
import sys
from hugr import tys
from hugr.tys import TypeBound

import guppylang.decorator as decorator


# Regular expression to match the `~~~~~^^^~~~` highlights that are printed in
# tracebacks from Python 3.11 onwards. We strip those out so we can use the same golden
# files for Python 3.10
TRACEBACK_HIGHLIGHT = re.compile(r" *~*\^\^*~*")


def run_error_test(file, capsys, snapshot):
    file = pathlib.Path(file)

    with pytest.raises(Exception) as exc_info:
        importlib.import_module(f"tests.error.{file.parent.name}.{file.name}")

    # Remove the importlib frames from the traceback by skipping beginning frames until
    # we end up in the executed file
    tb = exc_info.tb
    while tb is not None and inspect.getfile(tb.tb_frame) != str(file):
        tb = tb.tb_next

    # Invoke except hook to print the exception to stderr
    sys.excepthook(exc_info.type, exc_info.value.with_traceback(tb), tb)

    err = capsys.readouterr().err
    err = err.replace(str(file), "$FILE")

    # If we're comparing tracebacks, strip the highlights that are only present for
    # Python 3.11+
    if err.startswith("Traceback (most recent call last):"):
        err = "\n".join(
            line
            for line in err.split("\n")
            if not TRACEBACK_HIGHLIGHT.fullmatch(line)
        )

    snapshot.snapshot_dir = str(file.parent)
    snapshot.assert_match(err, file.with_suffix(".err").name)


@decorator.guppy.type(
    tys.Opaque(extension="", id="", args=[], bound=TypeBound.Copyable)
)
class NonBool:
    pass



from guppylang.decorator import guppy
from guppylang.std.builtins import array, owned
from guppylang.std.quantum import qubit


@guppy.declare
def foo(q: qubit, qs: array[qubit, 42] @owned) -> array[qubit, 42]: ...


@guppy
def main(qs: array[qubit, 42] @owned) -> array[qubit, 42]:
    return foo(qs[0], qs)


main.compile()



from guppylang.decorator import guppy

S = guppy.type_var("S")
T = guppy.type_var("T")
U = guppy.type_var("U")


@guppy.declare
def foo(x: S, y: T, z: U) -> None:
    ...


@guppy
def main() -> None:
    foo[int, int]


main.compile()



from guppylang.decorator import guppy
from guppylang.std.quantum import qubit
from guppylang.std.builtins import owned


@guppy.declare
def bar(q: qubit) -> int: ...


@guppy
def foo(n: int, q: qubit @owned) -> list[int]:
    return [bar(q) for _ in range(n)]


foo.compile()



from typing import no_type_check
from guppylang.decorator import guppy
from guppylang.std.option import Option, nothing, some
from guppylang.std.quantum import qubit  # noqa: TCH001


def test_none(run_int_fn):
    @guppy
    @no_type_check
    def main() -> int:
        x: Option[int] = nothing()
        is_none = 10 if x.is_nothing() else 0
        is_some = 1 if x.is_some() else 0
        return is_none + is_some

    run_int_fn(main, expected=10)


def test_some_unwrap(run_int_fn):
    @guppy
    @no_type_check
    def main() -> int:
        x: Option[int] = some(42)
        is_none = 1 if x.is_nothing() else 0
        is_some = x.unwrap() if x.is_some() else 0
        return is_none + is_some

    run_int_fn(main, expected=42)


def test_nothing_unwrap(run_int_fn):
    @guppy
    @no_type_check
    def main() -> int:
        x: Option[qubit] = nothing()
        x.unwrap_nothing()  # linearity error without this line
        return 1

    run_int_fn(main, expected=1)


def test_take(run_int_fn):
    @guppy
    @no_type_check
    def main() -> int:
        x: Option[int] = some(42)
        y = x.take().unwrap()
        is_none = 1 if x.is_nothing() else 0
        return y + is_none

    run_int_fn(main, expected=43)



from guppylang.std.builtins import result
from tests.util import compile_guppy

@compile_guppy
def foo(y: bool) -> None:
    # each tick or cross is 3 bytes. The cross sends the tag over the limit.
    result("✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅❌", y)



"""Tests validating the files in the `examples` directory."""

import pytest
from pathlib import Path

notebook_files = list(
    (Path(__file__).parent.parent.parent / "examples").glob("*.ipynb")
)


@pytest.mark.parametrize("notebook", notebook_files)
def test_example_notebooks(nb_regression, notebook: Path):
    nb_regression.diff_ignore += (
        "/metadata/language_info/version",
        "/cells/*/outputs/*/data/image/png",
    )
    nb_regression.check(notebook)


def test_demo_notebook(nb_regression):
    nb_regression.diff_ignore += ("/metadata/language_info/version",)
    nb_regression.check("tests/integration/notebooks/demo.ipynb")


def test_comptime_notebook(nb_regression):
    nb_regression.diff_ignore += ("/metadata/language_info/version",)
    nb_regression.check("tests/integration/notebooks/comptime.ipynb")


def test_misc_notebook_tests(nb_regression):
    nb_regression.diff_ignore += ("/metadata/language_info/version",)
    nb_regression.check("tests/integration/notebooks/misc_notebook_tests.ipynb")



from pytket import Circuit

from guppylang.decorator import guppy
from guppylang.std.quantum import qubit

circ = Circuit(2)
circ.X(0)
circ.Y(1)

@guppy.pytket(circ)
def guppy_circ(q: qubit) -> None: ...

@guppy
def foo(q: qubit) -> None:
    guppy_circ(q)


foo.compile()


from guppylang.decorator import guppy
from guppylang.std.builtins import owned
from guppylang.std.quantum import qubit


@guppy.declare
def foo(q1: qubit, q2: qubit @owned) -> qubit: ...


@guppy
def test(q1: qubit @owned, q2: qubit @owned) -> tuple[qubit, qubit]:
    q1 = foo(q1, q2)
    return q1, q2


test.compile()



from abc import ABC, abstractmethod
from collections.abc import Sequence
from dataclasses import dataclass, field
from enum import Enum, Flag, auto
from functools import cached_property, total_ordering
from typing import TYPE_CHECKING, ClassVar, TypeAlias, cast

import hugr.std.float
import hugr.std.int
from hugr import tys as ht

from guppylang_internals.error import InternalGuppyError
from guppylang_internals.tys.arg import Argument, ConstArg, TypeArg
from guppylang_internals.tys.common import (
    QuantifiedToHugrContext,
    ToHugr,
    ToHugrContext,
    Transformable,
    Transformer,
    Visitor,
)
from guppylang_internals.tys.const import Const, ConstValue, ExistentialConstVar
from guppylang_internals.tys.param import ConstParam, Parameter
from guppylang_internals.tys.var import BoundVar, ExistentialVar

if TYPE_CHECKING:
    from guppylang_internals.definition.struct import CheckedStructDef, StructField
    from guppylang_internals.definition.ty import OpaqueTypeDef
    from guppylang_internals.tys.subst import Inst, PartialInst, Subst


@dataclass(frozen=True)
class TypeBase(ToHugr[ht.Type], Transformable["Type"], ABC):
    """Abstract base class for all Guppy types.

    Note that all subclasses are expected to be immutable.
    """

    @cached_property
    @abstractmethod
    def copyable(self) -> bool:
        """Whether objects of this type can be implicitly copied."""

    @cached_property
    @abstractmethod
    def droppable(self) -> bool:
        """Whether objects of this type can be dropped."""

    @property
    def linear(self) -> bool:
        """Whether this type should be treated linearly."""
        return not self.copyable and not self.droppable

    @property
    def affine(self) -> bool:
        """Whether this type should be treated in an affine way."""
        return not self.copyable and self.droppable

    @cached_property
    @abstractmethod
    def hugr_bound(self) -> ht.TypeBound:
        """The Hugr bound of this type, i.e. `Any`, `Copyable`, or `Equatable`.

        This needs to be specified explicitly, since opaque nonlinear types in a Hugr
        extension could be either declared as copyable or equatable. If we don't get the
        bound exactly right during serialisation, the Hugr validator will complain.
        """

    @abstractmethod
    def cast(self) -> "Type":
        """Casts an implementor of `TypeBase` into a `Type`.

        This enforces that all implementors of `TypeBase` can be embedded into the
        `Type` union type.
        """

    @cached_property
    def unsolved_vars(self) -> set[ExistentialVar]:
        """The existential type variables contained in this type."""
        return set()

    def substitute(self, subst: "Subst") -> "Type":
        """Substitutes existential variables in this type."""
        from guppylang_internals.tys.subst import Substituter

        return self.transform(Substituter(subst))

    def to_arg(self) -> TypeArg:
        """Wraps this constant into a type argument."""
        return TypeArg(self.cast())

    def __str__(self) -> str:
        """Returns a human-readable representation of the type."""
        from guppylang_internals.tys.printing import TypePrinter

        # We use a custom printer that takes care of inserting parentheses and choosing
        # unique names
        return TypePrinter().visit(cast(Type, self))


@dataclass(frozen=True)
class ParametrizedTypeBase(TypeBase, ABC):
    """Abstract base class for types that depend on parameters.

    For example, `list`, `tuple`, etc. require arguments in order to be turned into a
    proper type.

    Note that all subclasses are expected to be immutable.
    """

    args: Sequence[Argument]

    def __post_init__(self) -> None:
        # Make sure that we don't have nested generic functions
        for arg in self.args:
            match arg:
                case TypeArg(ty=FunctionType(parametrized=True)):
                    raise InternalGuppyError(
                        "Tried to construct a higher-rank polymorphic type!"
                    )

    @property
    @abstractmethod
    def intrinsically_copyable(self) -> bool:
        """Whether this type is copyable, independent of the arguments.

        For example, a parametrized struct containing a qubit is never copyable, even if
        all its arguments are.
        """

    @cached_property
    def copyable(self) -> bool:
        """Whether this type should be treated as copyable."""
        # Either an argument isn't a type argument, or it must be copyable.
        return self.intrinsically_copyable and all(
            not isinstance(arg, TypeArg) or arg.ty.copyable for arg in self.args
        )

    @property
    @abstractmethod
    def intrinsically_droppable(self) -> bool:
        """Whether this type is droppable, independent of the arguments.

        For example, a parametrized struct containing a qubit is never droppable, even
        if all its arguments are.
        """

    @cached_property
    def droppable(self) -> bool:
        """Whether this type should be treated as copyable."""
        # Either an argument isn't a type argument, or it must be droppable.
        return self.intrinsically_droppable and all(
            not isinstance(arg, TypeArg) or arg.ty.droppable for arg in self.args
        )

    @cached_property
    def unsolved_vars(self) -> set[ExistentialVar]:
        """The existential type variables contained in this type."""
        return set().union(*(arg.unsolved_vars for arg in self.args))

    @cached_property
    def hugr_bound(self) -> ht.TypeBound:
        """The Hugr bound of this type, i.e. `Any`, `Copyable`, or `Equatable`."""
        if self.linear:
            return ht.TypeBound.Linear
        return ht.TypeBound.join(
            *(arg.ty.hugr_bound for arg in self.args if isinstance(arg, TypeArg))
        )

    def visit(self, visitor: Visitor) -> None:
        """Accepts a visitor on this type."""
        if not visitor.visit(self):
            for arg in self.args:
                visitor.visit(arg)


@dataclass(frozen=True)
class BoundTypeVar(TypeBase, BoundVar):
    """Bound type variable, referencing a parameter of kind `Type`.

    For example, in the function type `forall T. list[T] -> T` we represent `T` as a
    `BoundTypeVar(idx=0)`.

    A bound type variables can be instantiated with a `TypeArg` argument.
    """

    copyable: bool
    droppable: bool

    @cached_property
    def hugr_bound(self) -> ht.TypeBound:
        """The Hugr bound of this type, i.e. `Any`, `Copyable`, or `Equatable`."""
        if self.linear:
            return ht.TypeBound.Linear
        # We're conservative and don't require equatability for non-linear variables.
        # This is fine since Guppy doesn't use the equatable feature anyways.
        return ht.TypeBound.Copyable

    def cast(self) -> "Type":
        """Casts an implementor of `TypeBase` into a `Type`."""
        return self

    def to_hugr(self, ctx: ToHugrContext) -> ht.Type:
        """Computes the Hugr representation of the type."""
        return ctx.type_var_to_hugr(self)

    def visit(self, visitor: Visitor) -> None:
        """Accepts a visitor on this type."""
        visitor.visit(self)

    def transform(self, transformer: Transformer) -> "Type":
        """Accepts a transformer on this type."""
        return transformer.transform(self) or self

    def __str__(self) -> str:
        """Returns a human-readable representation of the type."""
        return self.display_name


@dataclass(frozen=True)
class ExistentialTypeVar(ExistentialVar, TypeBase):
    """Existential type variable.

    For example, the empty list literal `[]` is typed as `list[?T]` where `?T` stands
    for an existential type variable.

    During type checking we try to solve all existential type variables and substitute
    them with concrete types.
    """

    copyable: bool
    droppable: bool

    @classmethod
    def fresh(
        cls, display_name: str, copyable: bool, droppable: bool
    ) -> "ExistentialTypeVar":
        return ExistentialTypeVar(
            display_name, next(cls._fresh_id), copyable, droppable
        )

    @cached_property
    def unsolved_vars(self) -> set[ExistentialVar]:
        """The existential type variables contained in this type."""
        return {self}

    @cached_property
    def hugr_bound(self) -> ht.TypeBound:
        """The Hugr bound of this type, i.e. `Any`, `Copyable`, or `Equatable`."""
        raise InternalGuppyError(
            "Tried to compute bound of unsolved existential type variable"
        )

    def cast(self) -> "Type":
        """Casts an implementor of `TypeBase` into a `Type`."""
        return self

    def to_hugr(self, ctx: ToHugrContext) -> ht.Type:
        """Computes the Hugr representation of the type."""
        raise InternalGuppyError(
            "Tried to convert unsolved existential type variable to Hugr"
        )

    def visit(self, visitor: Visitor) -> None:
        """Accepts a visitor on this type."""
        visitor.visit(self)

    def transform(self, transformer: Transformer) -> "Type":
        """Accepts a transformer on this type."""
        return transformer.transform(self) or self


@dataclass(frozen=True)
class NoneType(TypeBase):
    """Type of `None`."""

    copyable: bool = field(default=True, init=True)
    droppable: bool = field(default=True, init=True)
    hugr_bound: ht.TypeBound = field(default=ht.TypeBound.Copyable, init=False)

    # Flag to avoid turning the type into a row when calling `type_to_row()`. This is
    # used to make sure that type vars instantiated to Nones are not broken up into
    # empty rows when generating a Hugr
    preserve: bool = field(default=False, compare=False)

    def cast(self) -> "Type":
        """Casts an implementor of `TypeBase` into a `Type`."""
        return self

    def to_hugr(self, ctx: ToHugrContext) -> ht.Tuple:
        """Computes the Hugr representation of the type."""
        return ht.Tuple()

    def visit(self, visitor: Visitor) -> None:
        """Accepts a visitor on this type."""
        visitor.visit(self)

    def transform(self, transformer: Transformer) -> "Type":
        """Accepts a transformer on this type."""
        return transformer.transform(self) or self


@dataclass(frozen=True)
class NumericType(TypeBase):
    """Numeric types like `int` and `float`."""

    kind: "Kind"

    @total_ordering
    class Kind(Enum):
        """The different kinds of numeric types."""

        Nat = auto()
        Int = auto()
        Float = auto()

        def __lt__(self, other: "NumericType.Kind") -> bool:
            return self.value < other.value

    INT_WIDTH: ClassVar[int] = 6

    @property
    def copyable(self) -> bool:
        """Whether objects of this type can be implicitly copied."""
        return True

    @property
    def droppable(self) -> bool:
        """Whether objects of this type can be dropped."""
        return True

    def cast(self) -> "Type":
        """Casts an implementor of `TypeBase` into a `Type`."""
        return self

    def to_hugr(self, ctx: ToHugrContext) -> ht.ExtType:
        """Computes the Hugr representation of the type."""
        match self.kind:
            case NumericType.Kind.Nat | NumericType.Kind.Int:
                return hugr.std.int.int_t(NumericType.INT_WIDTH)
            case NumericType.Kind.Float:
                return hugr.std.float.FLOAT_T

    @property
    def hugr_bound(self) -> ht.TypeBound:
        """The Hugr bound of this type, i.e. `Any` or `Copyable`"""
        return ht.TypeBound.Copyable

    def visit(self, visitor: Visitor) -> None:
        """Accepts a visitor on this type."""
        visitor.visit(self)

    def transform(self, transformer: Transformer) -> "Type":
        """Accepts a transformer on this type."""
        return transformer.transform(self) or self


class InputFlags(Flag):
    """Flags that can be set on inputs of function types.

    In the future, we could add  additional flags like `Frozen`.
    """

    NoFlags = 0
    Inout = auto()
    Owned = auto()
    Comptime = auto()


@dataclass(frozen=True)
class FuncInput:
    """A single input of a function type."""

    ty: "Type"
    flags: InputFlags


@dataclass(frozen=True, init=False)
class FunctionType(ParametrizedTypeBase):
    """Type of (potentially generic) functions."""

    inputs: Sequence[FuncInput]
    output: "Type"
    params: Sequence[Parameter]
    input_names: Sequence[str] | None
    comptime_args: Sequence[ConstArg]

    args: Sequence[Argument] = field(init=False)
    copyable: bool = field(default=True, init=True)
    droppable: bool = field(default=True, init=True)
    intrinsically_copyable: bool = field(default=True, init=True)
    intrinsically_droppable: bool = field(default=True, init=True)
    hugr_bound: ht.TypeBound = field(default=ht.TypeBound.Copyable, init=False)

    def __init__(
        self,
        inputs: Sequence[FuncInput],
        output: "Type",
        input_names: Sequence[str] | None = None,
        params: Sequence[Parameter] | None = None,
        comptime_args: Sequence[ConstArg] | None = None,
    ) -> None:
        # We need a custom __init__ to set the args
        args: list[Argument] = [TypeArg(inp.ty) for inp in inputs]
        args.append(TypeArg(output))

        # If no explicit comptime args are provided, assume that all of them are bound
        params = params or []
        if comptime_args is None:
            comptime_args = [
                param.to_bound()
                for param in params
                if isinstance(param, ConstParam) and param.from_comptime_arg
            ]
        args += comptime_args

        object.__setattr__(self, "args", args)
        object.__setattr__(self, "comptime_args", comptime_args)
        object.__setattr__(self, "inputs", inputs)
        object.__setattr__(self, "output", output)
        object.__setattr__(self, "input_names", input_names or [])
        object.__setattr__(self, "params", params)

    @property
    def parametrized(self) -> bool:
        """Whether the function is parametrized."""
        return len(self.params) > 0

    def cast(self) -> "Type":
        """Casts an implementor of `TypeBase` into a `Type`."""
        return self

    def to_hugr(self, ctx: ToHugrContext) -> ht.FunctionType:
        """Computes the Hugr representation of the type."""
        if self.parametrized:
            raise InternalGuppyError(
                "Tried to convert parametrised function type to Hugr. Use "
                "`to_hugr_poly` instead"
            )
        return self._to_hugr_function_type(ctx)

    def to_hugr_poly(self, ctx: ToHugrContext) -> ht.PolyFuncType:
        """Computes the Hugr `PolyFuncType` representation of the type."""
        # Function body needs to be translated in a new context where the variables are
        # bound to the quantifier.
        inner_ctx = QuantifiedToHugrContext(self.params)
        func_ty = self._to_hugr_function_type(inner_ctx)
        return ht.PolyFuncType(
            params=[p.to_hugr(ctx) for p in self.params], body=func_ty
        )

    def _to_hugr_function_type(self, ctx: ToHugrContext) -> ht.FunctionType:
        """Helper method to compute the Hugr `FunctionType` representation of the type.

        The resulting `FunctionType` can then be embedded into a Hugr `Type` or a Hugr
        `PolyFuncType`.
        """
        ins = [
            inp.ty.to_hugr(ctx)
            for inp in self.inputs
            # Comptime inputs are turned into generic args, so are not included here
            if InputFlags.Comptime not in inp.flags
        ]
        outs = [
            *(t.to_hugr(ctx) for t in type_to_row(self.output)),
            # We might have additional borrowed args that will be also outputted
            *(
                inp.ty.to_hugr(ctx)
                for inp in self.inputs
                if InputFlags.Inout in inp.flags
            ),
        ]
        return ht.FunctionType(input=ins, output=outs)

    def visit(self, visitor: Visitor) -> None:
        """Accepts a visitor on this type."""
        if not visitor.visit(self):
            for inp in self.inputs:
                visitor.visit(inp)
            visitor.visit(self.output)
            for param in self.params:
                visitor.visit(param)

    def transform(self, transformer: Transformer) -> "Type":
        """Accepts a transformer on this type."""
        return transformer.transform(self) or FunctionType(
            [
                FuncInput(inp.ty.transform(transformer), inp.flags)
                for inp in self.inputs
            ],
            self.output.transform(transformer),
            self.input_names,
            self.params,
        )

    def instantiate_partial(self, args: "PartialInst") -> "FunctionType":
        """Instantiates a subset of the function parameters with concrete types."""
        from guppylang_internals.tys.subst import Instantiator

        assert len(args) == len(self.params)

        full_inst: list[Argument] = []
        remaining_params: list[Parameter] = []
        for param, arg in zip(self.params, args, strict=True):
            # If no instantiation for this param is provided, it should stay around.
            # However, we have to down-shift the de Bruijn index.
            if arg is None:
                param = param.with_idx(len(remaining_params))
                remaining_params.append(param)
                arg = param.to_bound()

            # Set the `preserve` flag for instantiated tuples and None
            if isinstance(arg, TypeArg):
                if isinstance(arg.ty, TupleType):
                    arg = TypeArg(TupleType(arg.ty.element_types, preserve=True))
                elif isinstance(arg.ty, NoneType):
                    arg = TypeArg(NoneType(preserve=True))
            full_inst.append(arg)

        inst = Instantiator(full_inst)
        return FunctionType(
            [FuncInput(inp.ty.transform(inst), inp.flags) for inp in self.inputs],
            self.output.transform(inst),
            self.input_names,
            remaining_params,
            # Comptime type arguments also need to be instantiated
            comptime_args=[
                cast(ConstArg, arg.transform(inst)) for arg in self.comptime_args
            ],
        )

    def instantiate(self, args: "Inst") -> "FunctionType":
        """Instantiates all function parameters with concrete types."""
        return self.instantiate_partial(args)

    def unquantified(self) -> tuple["FunctionType", Sequence[ExistentialVar]]:
        """Instantiates all parameters with existential variables."""
        exs = [param.to_existential() for param in self.params]
        return self.instantiate([arg for arg, _ in exs]), [var for _, var in exs]


@dataclass(frozen=True, init=False)
class TupleType(ParametrizedTypeBase):
    """Type of tuples."""

    element_types: Sequence["Type"]

    # Flag to avoid turning the tuple into a row when calling `type_to_row()`. This is
    # used to make sure that type vars instantiated to tuples are not broken up into
    # rows when generating a Hugr
    preserve: bool = field(default=False, compare=False)

    def __init__(self, element_types: Sequence["Type"], preserve: bool = False) -> None:
        # We need a custom __init__ to set the args
        args = [TypeArg(ty) for ty in element_types]
        object.__setattr__(self, "args", args)
        object.__setattr__(self, "element_types", element_types)
        object.__setattr__(self, "preserve", preserve)

    @property
    def intrinsically_copyable(self) -> bool:
        """Whether objects of this type can be implicitly copied."""
        return True

    @property
    def intrinsically_droppable(self) -> bool:
        """Whether objects of this type can be dropped."""
        return True

    def cast(self) -> "Type":
        """Casts an implementor of `TypeBase` into a `Type`."""
        return self

    def to_hugr(self, ctx: ToHugrContext) -> ht.Tuple:
        """Computes the Hugr representation of the type."""
        return ht.Tuple(*row_to_hugr(self.element_types, ctx))

    def transform(self, transformer: Transformer) -> "Type":
        """Accepts a transformer on this type."""
        return transformer.transform(self) or TupleType(
            [ty.transform(transformer) for ty in self.element_types], self.preserve
        )


@dataclass(frozen=True, init=False)
class SumType(ParametrizedTypeBase):
    """Type of sums.

    Note that this type is only used internally when constructing the Hugr. Users cannot
    write down this type.
    """

    element_types: Sequence["Type"]

    def __init__(self, element_types: Sequence["Type"]) -> None:
        # We need a custom __init__ to set the args
        args = [TypeArg(ty) for ty in element_types]
        object.__setattr__(self, "args", args)
        object.__setattr__(self, "element_types", element_types)

    @property
    def intrinsically_copyable(self) -> bool:
        """Whether objects of this type can be implicitly copied."""
        return True

    @property
    def intrinsically_droppable(self) -> bool:
        """Whether objects of this type can be dropped."""
        return True

    def cast(self) -> "Type":
        """Casts an implementor of `TypeBase` into a `Type`."""
        return self

    def to_hugr(self, ctx: ToHugrContext) -> ht.Sum:
        """Computes the Hugr representation of the type."""
        rows = [type_to_row(ty) for ty in self.element_types]
        if all(len(row) == 0 for row in rows):
            return ht.UnitSum(size=len(rows))
        elif len(rows) == 1:
            return ht.Tuple(*row_to_hugr(rows[0], ctx))
        else:
            return ht.Sum(variant_rows=rows_to_hugr(rows, ctx))

    def transform(self, transformer: Transformer) -> "Type":
        """Accepts a transformer on this type."""
        return transformer.transform(self) or SumType(
            [ty.transform(transformer) for ty in self.element_types]
        )


@dataclass(frozen=True)
class OpaqueType(ParametrizedTypeBase):
    """Type that is directly backed by a Hugr opaque type.

    For example, many builtin types like `int`, `float`, `list` etc. are directly backed
    by a Hugr extension.
    """

    defn: "OpaqueTypeDef"

    @property
    def intrinsically_copyable(self) -> bool:
        """Whether objects of this type can be implicitly copied."""
        return not self.defn.never_copyable

    @property
    def intrinsically_droppable(self) -> bool:
        """Whether objects of this type can be dropped."""
        return not self.defn.never_droppable

    @property
    def hugr_bound(self) -> ht.TypeBound:
        """The Hugr bound of this type, i.e. `Any`, `Copyable`, or `Equatable`."""
        if self.defn.bound is not None:
            return self.defn.bound
        return super().hugr_bound

    def cast(self) -> "Type":
        """Casts an implementor of `TypeBase` into a `Type`."""
        return self

    def to_hugr(self, ctx: ToHugrContext) -> ht.Type:
        """Computes the Hugr representation of the type."""
        return self.defn.to_hugr(self.args, ctx)

    def transform(self, transformer: Transformer) -> "Type":
        """Accepts a transformer on this type."""
        return transformer.transform(self) or OpaqueType(
            [arg.transform(transformer) for arg in self.args], self.defn
        )


@dataclass(frozen=True)
class StructType(ParametrizedTypeBase):
    """A struct type."""

    defn: "CheckedStructDef"

    @cached_property
    def fields(self) -> list["StructField"]:
        """The fields of this struct type."""
        from guppylang_internals.definition.struct import StructField
        from guppylang_internals.tys.subst import Instantiator

        inst = Instantiator(self.args)
        return [StructField(f.name, f.ty.transform(inst)) for f in self.defn.fields]

    @cached_property
    def field_dict(self) -> "dict[str, StructField]":
        """Mapping from names to fields of this struct type."""
        return {field.name: field for field in self.fields}

    @cached_property
    def intrinsically_copyable(self) -> bool:
        """Whether objects of this type can be  implicitly copied."""
        return all(f.ty.copyable for f in self.fields)

    @cached_property
    def intrinsically_droppable(self) -> bool:
        """Whether objects of this type can be dropped."""
        return all(f.ty.droppable for f in self.fields)

    def cast(self) -> "Type":
        """Casts an implementor of `TypeBase` into a `Type`."""
        return self

    def to_hugr(self, ctx: ToHugrContext) -> ht.Tuple:
        """Computes the Hugr representation of the type."""
        return ht.Tuple(*(f.ty.to_hugr(ctx) for f in self.fields))

    def transform(self, transformer: Transformer) -> "Type":
        """Accepts a transformer on this type."""
        return transformer.transform(self) or StructType(
            [arg.transform(transformer) for arg in self.args], self.defn
        )


#: The type of parametrized Guppy types.
ParametrizedType: TypeAlias = (
    FunctionType | TupleType | SumType | OpaqueType | StructType
)

#: The type of Guppy types.
#:
#: This is a type alias for a union of all Guppy types defined in this module. This
#: models an algebraic data type and enables exhaustiveness checking in pattern matches
#: etc.
#:
#: This might become obsolete in case the @sealed decorator is added:
#:   * https://peps.python.org/pep-0622/#sealed-classes-as-algebraic-data-types
#:   * https://github.com/johnthagen/sealed-typing-pep
Type: TypeAlias = (
    BoundTypeVar | ExistentialTypeVar | NumericType | NoneType | ParametrizedType
)

#: An immutable row of Guppy types.
TypeRow: TypeAlias = Sequence[Type]


def row_to_type(row: TypeRow) -> Type:
    """Turns a row of types into a single type by packing into a tuple."""
    if len(row) == 0:
        return NoneType()
    elif len(row) == 1:
        return row[0]
    else:
        return TupleType(row)


def type_to_row(ty: Type) -> TypeRow:
    """Turns a type into a row of types by unpacking top-level tuples."""
    if isinstance(ty, NoneType) and not ty.preserve:
        return []
    if isinstance(ty, TupleType) and not ty.preserve:
        return ty.element_types
    return [ty]


def row_to_hugr(row: TypeRow, ctx: ToHugrContext) -> ht.TypeRow:
    """Computes the Hugr representation of a type row."""
    return [ty.to_hugr(ctx) for ty in row]


def rows_to_hugr(rows: Sequence[TypeRow], ctx: ToHugrContext) -> list[ht.TypeRow]:
    """Computes the Hugr representation of a sequence of rows."""
    return [row_to_hugr(row, ctx) for row in rows]


def unify(s: Type | Const, t: Type | Const, subst: "Subst | None") -> "Subst | None":
    """Computes a most general unifier for two types or constants.

    Return a substitutions `subst` such that `s[subst] == t[subst]` or `None` if this
    not possible.
    """
    # Make sure that s and t are either both constants or both types
    assert isinstance(s, TypeBase) == isinstance(t, TypeBase)
    if subst is None:
        return None
    match s, t:
        case ExistentialVar(id=s_id), ExistentialVar(id=t_id) if s_id == t_id:
            return subst
        case ExistentialTypeVar() | ExistentialConstVar() as s_var, t:
            return _unify_var(s_var, t, subst)
        case s, ExistentialTypeVar() | ExistentialConstVar() as t_var:
            return _unify_var(t_var, s, subst)
        case BoundVar(idx=s_idx), BoundVar(idx=t_idx) if s_idx == t_idx:
            return subst
        case ConstValue(value=c_value), ConstValue(value=d_value) if c_value == d_value:
            return subst
        case NumericType(kind=s_kind), NumericType(kind=t_kind) if s_kind == t_kind:
            return subst
        case NoneType(), NoneType():
            return subst
        case FunctionType() as s, FunctionType() as t if s.params == t.params:
            if len(s.inputs) != len(t.inputs):
                return None
            for a, b in zip(s.inputs, t.inputs, strict=True):
                if a.ty.linear and b.ty.linear and a.flags != b.flags:
                    return None
            return _unify_args(s, t, subst)
        case TupleType() as s, TupleType() as t:
            return _unify_args(s, t, subst)
        case SumType() as s, SumType() as t:
            return _unify_args(s, t, subst)
        case OpaqueType() as s, OpaqueType() as t if s.defn == t.defn:
            return _unify_args(s, t, subst)
        case StructType() as s, StructType() as t if s.defn == t.defn:
            return _unify_args(s, t, subst)
        case _:
            return None


def _unify_var(
    var: ExistentialTypeVar | ExistentialConstVar, t: Type | Const, subst: "Subst"
) -> "Subst | None":
    """Helper function for unification of type or const variables."""
    if var in subst:
        return unify(subst[var], t, subst)
    if isinstance(t, ExistentialTypeVar) and t in subst:
        return unify(var, subst[t], subst)
    if var in t.unsolved_vars:
        return None
    return {var: t, **subst}


def _unify_args(
    s: ParametrizedType, t: ParametrizedType, subst: "Subst"
) -> "Subst | None":
    """Helper function for unification of type arguments of parametrised types."""
    if len(s.args) != len(t.args):
        return None
    for sa, ta in zip(s.args, t.args, strict=True):
        match sa, ta:
            case TypeArg(ty=sa_ty), TypeArg(ty=ta_ty):
                res = unify(sa_ty, ta_ty, subst)
                if res is None:
                    return None
                subst = res
            case ConstArg(const=sa_const), ConstArg(const=ta_const):
                res = unify(sa_const, ta_const, subst)
                if res is None:
                    return None
                subst = res
            case _:
                return None
    return subst


### Helpers for working with tuples of functions


def parse_function_tensor(ty: TupleType) -> list[FunctionType] | None:
    """Parses a nested tuple of function types into a flat list of functions."""
    result = []
    for el in ty.element_types:
        if isinstance(el, FunctionType):
            result.append(el)
        elif isinstance(el, TupleType):
            funcs = parse_function_tensor(el)
            if funcs:
                result.extend(funcs)
            else:
                return None
    return result


def function_tensor_signature(tys: list[FunctionType]) -> FunctionType:
    """Compute the combined function signature of a list of functions"""
    inputs: list[FuncInput] = []
    outputs: list[Type] = []
    for fun_ty in tys:
        assert not fun_ty.parametrized
        inputs.extend(fun_ty.inputs)
        outputs.extend(type_to_row(fun_ty.output))
    return FunctionType(inputs, row_to_type(outputs))



from dataclasses import dataclass
from typing import ClassVar

from guppylang_internals.diagnostic import Error
from guppylang_internals.tys.ty import Type


@dataclass(frozen=True)
class WasmError(Error):
    title: ClassVar[str] = "WASM signature error"


@dataclass(frozen=True)
class FirstArgNotModule(WasmError):
    span_label: ClassVar[str] = (
        "First argument to WASM function should be a reference to a WASM module."
        " Found `{ty}` instead"
    )
    ty: Type


@dataclass(frozen=True)
class UnWasmableType(WasmError):
    span_label: ClassVar[str] = (
        "WASM function signature contained an unsupported type: `{ty}`"
    )
    ty: Type


@dataclass(frozen=True)
class WasmTypeConversionError(Error):
    title: ClassVar[str] = "Can't convert type to WASM"
    span_label: ClassVar[str] = "`{thing}` cannot be converted to WASM"
    ty: Type



from guppylang.decorator import guppy
from guppylang.std.builtins import owned
from guppylang.std.quantum import qubit


@guppy.struct
class MyStruct:
    q: qubit


@guppy
def foo(s: MyStruct @owned) -> tuple[qubit, qubit]:
    t = s
    return s.q, t.q


foo.compile()



class PyDiagnostic:
    def __init__(
        self,
        message: str,
        code: str | None,
        severity: str,
        source: str | None,
        spans: list[tuple[int, int, str | None]],
        help_text: str | None,
    ) -> None: ...

def render_report(diagnostic: PyDiagnostic) -> str: ...
def guppy_to_miette(
    title: str,
    level: str,
    source_text: str | None,
    spans: list[tuple[int, int, str | None]],
    message: str | None,
    help_text: str | None,
) -> PyDiagnostic: ...



from collections.abc import Callable, Sequence

from hugr import ops
from hugr import tys as ht

from guppylang_internals.error import InternalGuppyError
from guppylang_internals.std._internal.compiler.tket_exts import FUTURES_EXTENSION
from guppylang_internals.tys.arg import Argument, TypeArg
from guppylang_internals.tys.common import ToHugrContext
from guppylang_internals.tys.subst import Inst


def future_to_hugr(type_args: Sequence[Argument], ctx: ToHugrContext) -> ht.Type:
    match type_args:
        case [TypeArg(ty)]:
            type_def = FUTURES_EXTENSION.get_type("Future")
            return type_def.instantiate([ht.TypeTypeArg(ty.to_hugr(ctx))])
        case _:
            raise InternalGuppyError("Invalid type args for Future type")


def future_op(
    op_name: str,
) -> Callable[[ht.FunctionType, Inst, ToHugrContext], ops.DataflowOp]:
    op_def = FUTURES_EXTENSION.get_op(op_name)

    def op(concrete: ht.FunctionType, args: Inst, ctx: ToHugrContext) -> ops.DataflowOp:
        return op_def.instantiate([arg.to_hugr(ctx) for arg in args], concrete)

    return op



from guppylang.decorator import guppy
from guppylang.std.builtins import owned
from guppylang.std.quantum import qubit


@guppy
def foo(q: qubit @owned) -> qubit:
    def bar() -> qubit:
        return q

    return q


foo.compile()



"""Mocks for a number of builtin functions to be used during tracing.

For example, the builtin `int(x)` function tries calling `x.__int__()` and raises a
`TypeError` if this call doesn't return an `int`. During tracing however, we also want
to allow `int(x)` if `x` is an abstract `GuppyObject`. To allow this, we need to mock
the builtins to avoid raising the `TypeError` in that case.
"""

import builtins
from collections.abc import Callable
from typing import Any

from guppylang_internals.tracing.object import GuppyObject, GuppyStructObject


def _mock_meta(cls: type) -> type:
    """Returns a metaclass that replicates the behaviour of the provided class.

    The only way to distinguishing a `_mock_meta(T)` from an actual `T` is via checking
    reference equality using the `is` operator.
    """

    class MockMeta(type):
        def __instancecheck__(self, instance: Any) -> bool:
            return cls.__instancecheck__(instance)

        def __subclasscheck__(self, subclass: type) -> bool:
            return cls.__subclasscheck__(subclass)

        def __eq__(self, other: object) -> Any:
            return other == cls

        def __ne__(self, other: object) -> Any:
            return other != cls

    MockMeta.__name__ = type.__name__
    MockMeta.__qualname__ = type.__qualname__
    return MockMeta


class float(builtins.float, metaclass=_mock_meta(builtins.float)):  # type: ignore[misc]
    def __new__(cls, x: Any = 0.0, /) -> Any:
        if isinstance(x, GuppyObject):
            return x.__float__()
        return builtins.float(x)


class int(builtins.int, metaclass=_mock_meta(builtins.int)):  # type: ignore[misc]
    def __new__(cls, x: Any = 0, /, **kwargs: Any) -> Any:
        if isinstance(x, GuppyObject):
            return x.__int__(**kwargs)
        return builtins.int(x, **kwargs)


def len(x: Any) -> Any:
    if isinstance(x, GuppyObject | GuppyStructObject):
        return x.__len__()
    return builtins.len(x)


def mock_builtins(f: Callable[..., Any]) -> None:
    f.__globals__.update({"float": float, "int": int, "len": len})



from collections.abc import Callable
from typing import Generic

from guppylang.decorator import guppy


def test_create(run_int_fn):
    @guppy.struct
    class S:
        x: int
        y: int

    @guppy.comptime
    def main(x: int) -> int:
        s = S(x, 2)
        return s.x + s.y

    run_int_fn(main, 42, args=[40])


def test_argument(run_int_fn):
    @guppy.struct
    class S:
        x: int
        y: int

    @guppy.comptime
    def foo(s: S) -> int:
        return s.x + s.y

    @guppy
    def main() -> int:
        return foo(S(40, 2))

    run_int_fn(main, 42)


def test_write(run_int_fn):
    @guppy.struct
    class S:
        x: int
        y: int

    @guppy.comptime
    def main(y: int) -> int:
        s = S(40, y)
        t = s
        t.y += 1
        return s.x + s.y

    run_int_fn(main, 42, args=[1])


def test_method(run_int_fn):
    @guppy.struct
    class S:
        x: int
        y: int

        @guppy.comptime
        def get_x(self: "S") -> int:
            return self.x

        @guppy
        def get_y(self: "S") -> int:
            return self.y

    @guppy.comptime
    def main(x: int, y: int) -> int:
        s = S(x, y)
        return s.get_x() + s.get_y()

    run_int_fn(main, 42, args=[40, 2])


def test_generic_nested(validate, run_float_fn_approx):
    S = guppy.type_var("S")
    T = guppy.type_var("T")

    @guppy.struct
    class StructA(Generic[T]):
        x: tuple[int, T]

    @guppy.struct
    class StructB(Generic[S, T]):
        x: S
        y: StructA[T]

    @guppy.comptime
    def foo(a: StructA[StructA[float]], b: StructB[bool, int]) -> float:
        flat_a = a.x[0] + a.x[1].x[0] + a.x[1].x[1]
        flat_b = b.y.x[0] + b.y.x[1]
        return flat_a + flat_b

    @guppy.comptime
    def bar(
        x1: int, x2: int, x3: float, x4: int, x5: int
    ) -> tuple[StructA[StructA[float]], StructB[bool, int]]:
        a = StructA((x1, StructA((x2, x3))))
        b = StructB(True, StructA((x4, x5)))
        return a, b

    @guppy
    def main() -> float:
        a, b = bar(1, 10, 100, 1000, 10000)
        return foo(a, b)

    run_float_fn_approx(main, 11111)


def test_load_constructor(validate):
    @guppy.struct
    class S:
        x: int

    @guppy.comptime
    def test() -> Callable[[int], S]:
        return S

    validate(test.compile())



from guppylang import guppy
from tests.util import compile_guppy
from guppylang.std.quantum import qubit
from guppylang.std.builtins import owned

@guppy.declare
def use(x: qubit @ owned) -> None: ...

@compile_guppy
def foo(b: bool) -> tuple[qubit, int]:
    t = (qubit(), 100)
    if b:
        use(t[0])
    return t


from guppylang.decorator import guppy
from guppylang.std.builtins import array

n = guppy.nat_var("n")


@guppy.declare
def foo(x: array[int, n]) -> None:
    ...


@guppy
def main(xs: array[int, 42]) -> None:
    foo[43](xs)


main.compile()



from typing import no_type_check
from guppylang.decorator import guppy
from guppylang.std.builtins import barrier
from guppylang.std.quantum import qubit


@guppy
@no_type_check
def main() -> qubit:
   q = qubit()
   barrier(q, q)
   return q


main.compile()



from guppylang.decorator import guppy
from guppylang.std.builtins import array
from guppylang.std.quantum import qubit


@guppy
def main(qs: array[qubit, 42]) -> tuple[qubit, array[qubit, 42]]:
    q = qs[0]
    return q, qs


main.compile()


from guppylang.decorator import guppy
from guppylang.std.either import Either, left, right
from guppylang.std.quantum import qubit  # noqa: TCH001


def test_left(run_int_fn):
    @guppy
    def main() -> int:
        x: Either[int, qubit] = left(100)
        is_left = 1 if x.is_left() else 0
        is_right = 10 if x.is_right() else 0
        return is_left + is_right + x.unwrap_left()

    run_int_fn(main, expected=101)


def test_right(run_int_fn):
    @guppy
    def main() -> int:
        x: Either[qubit, int] = right(100)
        is_left = 1 if x.is_left() else 0
        is_right = 10 if x.is_right() else 0
        return is_left + is_right + x.unwrap_right()

    run_int_fn(main, expected=110)


def test_to_option(run_int_fn):
    @guppy
    def main() -> int:
        l_val: Either[int, float] = left(1)
        r_val: Either[float, int] = right(10)
        l_val.try_into_right().unwrap_nothing()
        r_val.try_into_left().unwrap_nothing()
        return l_val.try_into_left().unwrap() + r_val.try_into_right().unwrap()

    run_int_fn(main, expected=11)



from guppylang.decorator import guppy
from guppylang.std.quantum import qubit
from guppylang.std.builtins import owned


@guppy
def foo(xs: list[int], q: qubit @owned) -> list[qubit]:
    return [q for x in xs]


foo.compile()



from guppylang.decorator import guppy
from guppylang.std.quantum import qubit
from guppylang.std.builtins import owned


@guppy
def foo(qs: list[qubit] @owned, xs: list[int]) -> list[int]:
    return [q for q in qs for q in xs]


foo.compile()



from guppylang.decorator import guppy
from guppylang.std.quantum import qubit
from guppylang.std.builtins import owned


@guppy.struct
class MyStruct:
    q1: qubit
    q2: qubit


@guppy
def foo(ss: list[MyStruct] @owned) -> list[qubit]:
    return [s.q1 for s in ss]


foo.compile()



from guppylang import guppy
from guppylang.std.builtins import comptime, nat

T = guppy.type_var("T")


@guppy
def foo(x: T, n: nat @comptime) -> None:
    pass


@guppy
def main() -> None:
    foo[int](42, 43)


main.compile()



from hugr import tys
from guppylang import guppy
from guppylang_internals.decorator import custom_type


@custom_type(tys.Tuple(), droppable=False)
class NonDroppable:
    pass


@guppy
def main[D: NonDroppable]() -> None:
    pass


main.compile()



from typing import Generic

from guppylang.decorator import guppy

T = guppy.type_var("T")


@guppy.struct
class MyStruct(Generic[T]):
    x: T
    y: T


@guppy
def main() -> None:
    MyStruct(0, False)


main.compile()



from guppylang.decorator import guppy
from guppylang.std.builtins import array, owned


@guppy.struct
class MyStruct:
    x: array[int, 1]


@guppy.declare
def use(arr: array[int, 1] @owned) -> None: ...


@guppy
def test(s: MyStruct) -> None:
    p = s.x
    use(s.x)


test.compile()


"""Constants are only loaded by execution compilers (not validation), test that."""

from guppylang.decorator import guppy
from guppylang.std.builtins import result, comptime


def test_basic_type(run_int_fn):
    @guppy.comptime
    def main() -> int:
        for x in [
            1,
            -1,
            0,
            1.2,
            -1.2,
            0.0,
            True,
            False,
            [1, 2, 3],
            [-1, -2, -3],
            [0, 0, 0],
            [1.2, 2.3, 3.4],
            [-1.2, -2.3, -3.4],
            [True, False, True],
        ]:
            result("x", x)
        return 0

    run_int_fn(main, 0)


def test_tuples(run_int_fn):
    # tuples cannot be `result` reported directly, so extract a value
    tup = (1, -1, 0, 1.2, -1.2, 0.0, True, False)

    @guppy
    def main() -> int:
        g_t = comptime(tup)
        x = g_t[0]
        result("x", x)
        return 0

    run_int_fn(main, 0)



"""Tests that non-comptime functions still get regular diagnostics, even if their
checking was invoked from a comptime function.

See https://github.com/CQCL/guppylang/issues/1097
"""

from guppylang.decorator import guppy


@guppy
def foo() -> int:
    return


@guppy.comptime
def main() -> int:
    return foo()


main.compile()



from guppylang.decorator import guppy
from guppylang.std.quantum import qubit


@guppy.declare
def new_qubit() -> qubit:
    ...


@guppy
def foo(b: bool) -> qubit:
    if b:
        q = new_qubit()
    else:
        q = new_qubit()
    q = new_qubit()
    return q


foo.compile()



from guppylang.decorator import guppy
from guppylang.std.builtins import owned
from guppylang.std.quantum import qubit


@guppy
def foo(q: qubit @owned) -> None:
    while True:
        pass


foo.compile()



from guppylang.decorator import guppy
from guppylang.std.quantum import qubit, measure


@guppy.struct
class MyStruct:
    q: qubit


@guppy
def foo(b: bool) -> bool:
    s = MyStruct(qubit())
    if b:
        return measure(s.q)
    return False


foo.compile()



import ast
import sys
from collections.abc import Sequence
from types import ModuleType

from guppylang_internals.ast_util import (
    AstNode,
    set_location_from,
    shift_loc,
)
from guppylang_internals.cfg.builder import is_comptime_expression
from guppylang_internals.checker.core import Context, Globals, Locals, PythonObject
from guppylang_internals.checker.errors.generic import ExpectedError, UnsupportedError
from guppylang_internals.definition.common import Definition
from guppylang_internals.definition.parameter import ParamDef
from guppylang_internals.definition.ty import TypeDef
from guppylang_internals.engine import ENGINE
from guppylang_internals.error import GuppyError
from guppylang_internals.tys.arg import Argument, ConstArg, TypeArg
from guppylang_internals.tys.builtin import CallableTypeDef, bool_type
from guppylang_internals.tys.const import ConstValue
from guppylang_internals.tys.errors import (
    CallableComptimeError,
    ComptimeArgShadowError,
    FlagNotAllowedError,
    FreeTypeVarError,
    HigherKindedTypeVarError,
    IllegalComptimeTypeArgError,
    InvalidCallableTypeError,
    InvalidFlagError,
    InvalidTypeArgError,
    InvalidTypeError,
    LinearComptimeError,
    LinearConstParamError,
    ModuleMemberNotFoundError,
    NonLinearOwnedError,
)
from guppylang_internals.tys.param import ConstParam, Parameter, TypeParam
from guppylang_internals.tys.subst import BoundVarFinder
from guppylang_internals.tys.ty import (
    FuncInput,
    FunctionType,
    InputFlags,
    NoneType,
    NumericType,
    TupleType,
    Type,
)


def arg_from_ast(
    node: AstNode,
    globals: Globals,
    param_var_mapping: dict[str, Parameter],
    allow_free_vars: bool = False,
) -> Argument:
    """Turns an AST expression into an argument."""
    from guppylang_internals.checker.cfg_checker import VarNotDefinedError

    # A single (possibly qualified) identifier
    if defn := _try_parse_defn(node, globals):
        return _arg_from_instantiated_defn(
            defn, [], globals, node, param_var_mapping, allow_free_vars
        )

    # An identifier referring to a quantified variable
    if isinstance(node, ast.Name):
        if node.id in param_var_mapping:
            return param_var_mapping[node.id].to_bound()
        raise GuppyError(VarNotDefinedError(node, node.id))

    # A parametrised type, e.g. `list[??]`
    if isinstance(node, ast.Subscript) and (
        defn := _try_parse_defn(node.value, globals)
    ):
        arg_nodes = (
            node.slice.elts if isinstance(node.slice, ast.Tuple) else [node.slice]
        )
        return _arg_from_instantiated_defn(
            defn, arg_nodes, globals, node, param_var_mapping, allow_free_vars
        )

    # We allow tuple types to be written as `(int, bool)`
    if isinstance(node, ast.Tuple):
        ty = TupleType(
            [
                type_from_ast(el, globals, param_var_mapping, allow_free_vars)
                for el in node.elts
            ]
        )
        return TypeArg(ty)

    # Literals
    if isinstance(node, ast.Constant):
        match node.value:
            # `None` is represented as a `ast.Constant` node with value `None`
            case None:
                return TypeArg(NoneType())
            case bool(v):
                return ConstArg(ConstValue(bool_type(), v))
            # Integer literals are turned into nat args.
            # TODO: To support int args, we need proper inference logic here
            #   See https://github.com/CQCL/guppylang/issues/1030
            case int(v) if v >= 0:
                nat_ty = NumericType(NumericType.Kind.Nat)
                return ConstArg(ConstValue(nat_ty, v))
            case float(v):
                float_ty = NumericType(NumericType.Kind.Float)
                return ConstArg(ConstValue(float_ty, v))
            # String literals are ignored for now since they could also be stringified
            # types.
            # TODO: To support string args, we need proper inference logic here
            #   See https://github.com/CQCL/guppylang/issues/1030
            case str(_):
                pass

    # Py-expressions can also be used to specify static numbers
    if comptime_expr := is_comptime_expression(node):
        from guppylang_internals.checker.expr_checker import eval_comptime_expr

        v = eval_comptime_expr(comptime_expr, Context(globals, Locals({}), {}))
        if isinstance(v, int):
            nat_ty = NumericType(NumericType.Kind.Nat)
            return ConstArg(ConstValue(nat_ty, v))
        else:
            raise GuppyError(IllegalComptimeTypeArgError(node, v))

    # Finally, we also support delayed annotations in strings
    if isinstance(node, ast.Constant) and isinstance(node.value, str):
        node = _parse_delayed_annotation(node.value, node)
        return arg_from_ast(node, globals, param_var_mapping, allow_free_vars)

    raise GuppyError(InvalidTypeArgError(node))


def _try_parse_defn(node: AstNode, globals: Globals) -> Definition | None:
    """Tries to parse a (possibly qualified) name into a global definition."""
    from guppylang.defs import GuppyDefinition
    from guppylang_internals.checker.cfg_checker import VarNotDefinedError

    match node:
        case ast.Name(id=x):
            if x not in globals:
                return None
            defn = globals[x]
            if isinstance(defn, PythonObject):
                return None
            return defn
        case ast.Attribute(value=ast.Name(id=module_name) as value, attr=x):
            if module_name not in globals:
                raise GuppyError(VarNotDefinedError(value, module_name))
            match globals[module_name]:
                case PythonObject(ModuleType() as module):
                    if x in module.__dict__:
                        val = module.__dict__[x]
                        if isinstance(val, GuppyDefinition):
                            return ENGINE.get_parsed(val.id)
                    raise GuppyError(
                        ModuleMemberNotFoundError(node, module.__name__, x)
                    )
                case _:
                    raise GuppyError(ExpectedError(value, "a module"))
        case _:
            return None


def _arg_from_instantiated_defn(
    defn: Definition,
    arg_nodes: list[ast.expr],
    globals: Globals,
    node: AstNode,
    param_var_mapping: dict[str, Parameter],
    allow_free_vars: bool = False,
) -> Argument:
    """Parses a globals definition with type args into an argument."""
    match defn:
        # Special case for the `Callable` type
        case CallableTypeDef():
            return TypeArg(
                _parse_callable_type(
                    arg_nodes, node, globals, param_var_mapping, allow_free_vars
                )
            )
        # Either a defined type (e.g. `int`, `bool`, ...)
        case TypeDef() as defn:
            args = [
                arg_from_ast(arg_node, globals, param_var_mapping, allow_free_vars)
                for arg_node in arg_nodes
            ]
            ty = defn.check_instantiate(args, node)
            return TypeArg(ty)
        # Or a parameter (e.g. `T`, `n`, ...)
        case ParamDef() as defn:
            # We don't allow parametrised variables like `T[int]`
            if arg_nodes:
                raise GuppyError(HigherKindedTypeVarError(node, defn))
            if defn.name not in param_var_mapping:
                if allow_free_vars:
                    param_var_mapping[defn.name] = defn.to_param(len(param_var_mapping))
                else:
                    raise GuppyError(FreeTypeVarError(node, defn))
            return param_var_mapping[defn.name].to_bound()
        case defn:
            err = ExpectedError(node, "a type", got=f"{defn.description} `{defn.name}`")
            raise GuppyError(err)


def _parse_delayed_annotation(ast_str: str, node: ast.Constant) -> ast.expr:
    """Parses a delayed type annotation in a string."""
    try:
        [stmt] = ast.parse(ast_str).body
        if not isinstance(stmt, ast.Expr):
            raise GuppyError(InvalidTypeError(node))
        set_location_from(stmt, loc=node)
        shift_loc(
            stmt,
            delta_lineno=node.lineno - 1,  # -1 since lines start at 1
            delta_col_offset=node.col_offset + 1,  # +1 to remove the `"`
        )
    except (SyntaxError, ValueError):
        raise GuppyError(InvalidTypeError(node)) from None
    else:
        return stmt.value


def _parse_callable_type(
    args: list[ast.expr],
    loc: AstNode,
    globals: Globals,
    param_var_mapping: dict[str, Parameter],
    allow_free_vars: bool = False,
) -> FunctionType:
    """Helper function to parse a `Callable[[<arguments>], <return type>]` type."""
    err = InvalidCallableTypeError(loc)
    if len(args) != 2:
        raise GuppyError(err)
    [inputs, output] = args
    if not isinstance(inputs, ast.List):
        raise GuppyError(err)
    inouts, output = parse_function_io_types(
        inputs.elts, output, None, loc, globals, param_var_mapping, allow_free_vars
    )
    return FunctionType(inouts, output)


def parse_function_io_types(
    input_nodes: list[ast.expr],
    output_node: ast.expr,
    input_names: list[str] | None,
    loc: AstNode,
    globals: Globals,
    param_var_mapping: dict[str, Parameter],
    allow_free_vars: bool = False,
) -> tuple[list[FuncInput], Type]:
    """Parses the inputs and output types of a function type.

    This function takes care of parsing annotations and any related checks.

    Returns the parsed input and output types.
    """
    inputs = []
    for i, inp in enumerate(input_nodes):
        ty, flags = type_with_flags_from_ast(
            inp, globals, param_var_mapping, allow_free_vars
        )
        if InputFlags.Owned in flags and ty.copyable:
            raise GuppyError(NonLinearOwnedError(loc, ty))
        if not ty.copyable and InputFlags.Owned not in flags:
            flags |= InputFlags.Inout
        if InputFlags.Comptime in flags:
            if input_names is None:
                raise GuppyError(CallableComptimeError(inp))
            name = input_names[i]

            # Make sure we're not shadowing a type variable with the same name that was
            # already used on the left. E.g
            #
            #    n = guppy.type_var("n")
            #    def foo(xs: array[int, n], n: nat @comptime)
            #
            # TODO: In principle we could lift this restriction by tracking multiple
            #  params referring to the same name in `param_var_mapping`, but not sure if
            #  this would be worth it...
            if name in param_var_mapping:
                raise GuppyError(ComptimeArgShadowError(inp, name))
            param_var_mapping[name] = ConstParam(
                len(param_var_mapping), name, ty, from_comptime_arg=True
            )

        inputs.append(FuncInput(ty, flags))
    output = type_from_ast(output_node, globals, param_var_mapping, allow_free_vars)
    return inputs, output


if sys.version_info >= (3, 12):

    def parse_parameter(node: ast.type_param, idx: int, globals: Globals) -> Parameter:
        """Parses a `Variable: Bound` generic type parameter declaration."""
        if isinstance(node, ast.TypeVarTuple | ast.ParamSpec):
            raise GuppyError(UnsupportedError(node, "Variadic generic parameters"))
        assert isinstance(node, ast.TypeVar)

        match node.bound:
            # No bound means it's a linear type parameter
            case None:
                return TypeParam(
                    idx, node.name, must_be_copyable=False, must_be_droppable=False
                )
            # Special `Copy` or `Drop` bounds for types
            case ast.Name(id="Copy"):
                return TypeParam(
                    idx, node.name, must_be_copyable=True, must_be_droppable=False
                )
            case ast.Name(id="Drop"):
                return TypeParam(
                    idx, node.name, must_be_copyable=False, must_be_droppable=True
                )
            # Copy and drop is annotated as `T: (Copy, Drop)`
            # TODO: Should we also allow `T: Copy + Drop`? Mypy would complain about it
            case ast.Tuple(elts=[ast.Name(id=id1), ast.Name(id=id2)]) if {id1, id2} == {
                "Copy",
                "Drop",
            }:
                return TypeParam(
                    idx, node.name, must_be_copyable=True, must_be_droppable=True
                )
            # Otherwise, it must be a const parameter
            case bound:
                # For now, we don't allow the types of const params to refer to previous
                # parameters, so we pass an empty dict as the `param_var_mapping`.
                # TODO: In the future we might want to allow stuff like
                #   `def foo[T, XS: array[T, 42]]` and so on
                ty = type_from_ast(bound, globals, {}, allow_free_vars=False)
                if not ty.copyable or not ty.droppable:
                    raise GuppyError(LinearConstParamError(bound, ty))

                # TODO: For now we can only do `nat` const args since they lower to
                #  Hugr bounded nats. Extend to arbitrary types via monomorphization.
                #  See https://github.com/CQCL/guppylang/issues/1008
                if ty != NumericType(NumericType.Kind.Nat):
                    raise GuppyError(
                        UnsupportedError(bound, f"`{ty}` generic parameters")
                    )
                return ConstParam(idx, node.name, ty)


_type_param = TypeParam(0, "T", False, False)


def type_with_flags_from_ast(
    node: AstNode,
    globals: Globals,
    param_var_mapping: dict[str, Parameter],
    allow_free_vars: bool = False,
) -> tuple[Type, InputFlags]:
    if isinstance(node, ast.BinOp) and isinstance(node.op, ast.MatMult):
        ty, flags = type_with_flags_from_ast(
            node.left, globals, param_var_mapping, allow_free_vars
        )
        match node.right:
            case ast.Name(id="owned"):
                if ty.copyable:
                    raise GuppyError(NonLinearOwnedError(node.right, ty))
                flags |= InputFlags.Owned
            case ast.Name(id="comptime"):
                flags |= InputFlags.Comptime
                if not ty.copyable or not ty.droppable:
                    raise GuppyError(LinearComptimeError(node.right, ty))
                # For now, we don't allow comptime annotations on generic inputs
                # TODO: In the future we might want to allow stuff like
                #  `def foo[T: (Copy, Discard](x: T @comptime)`.
                #   Also see the todo in `parse_parameter`.
                var_finder = BoundVarFinder()
                ty.visit(var_finder)
                if var_finder.bound_vars:
                    raise GuppyError(
                        UnsupportedError(node.left, "Generic comptime arguments")
                    )
            case _:
                raise GuppyError(InvalidFlagError(node.right))
        return ty, flags
    # We also need to handle the case that this could be a delayed string annotation
    elif isinstance(node, ast.Constant) and isinstance(node.value, str):
        node = _parse_delayed_annotation(node.value, node)
        return type_with_flags_from_ast(
            node, globals, param_var_mapping, allow_free_vars
        )
    else:
        # Parse an argument and check that it's valid for a `TypeParam`
        arg = arg_from_ast(node, globals, param_var_mapping, allow_free_vars)
        tyarg = _type_param.check_arg(arg, node)
        return tyarg.ty, InputFlags.NoFlags


def type_from_ast(
    node: AstNode,
    globals: Globals,
    param_var_mapping: dict[str, Parameter],
    allow_free_vars: bool = False,
) -> Type:
    """Turns an AST expression into a Guppy type."""
    ty, flags = type_with_flags_from_ast(
        node, globals, param_var_mapping, allow_free_vars
    )
    if flags != InputFlags.NoFlags:
        assert InputFlags.Inout not in flags  # Users shouldn't be able to set this
        raise GuppyError(FlagNotAllowedError(node))
    return ty


def type_row_from_ast(
    node: ast.expr, globals: "Globals", allow_free_vars: bool = False
) -> Sequence[Type]:
    """Turns an AST expression into a Guppy type row.

    This is needed to interpret the return type annotation of functions.
    """
    # The return type `-> None` is represented in the ast as `ast.Constant(value=None)`
    if isinstance(node, ast.Constant) and node.value is None:
        return []
    ty = type_from_ast(node, globals, {}, allow_free_vars)
    if isinstance(ty, TupleType):
        return ty.element_types
    else:
        return [ty]



from guppylang.decorator import guppy
from guppylang.std.angles import angle, pi

from hugr.std.int import IntVal


def test_int(run_int_fn):
    @guppy.comptime
    def pos(x: int) -> int:
        return +x

    @guppy.comptime
    def neg(x: int) -> int:
        return -x

    @guppy.comptime
    def add(x: int, y: int) -> int:
        return 1 + (x + (y + 2))

    @guppy.comptime
    def sub(x: int, y: int) -> int:
        return 1 - (x - (y - 2))

    @guppy.comptime
    def mul(x: int, y: int) -> int:
        return 1 * (x * (y * 2))

    @guppy.comptime
    def div(x: int, y: int) -> int:
        return 100 // (x // (y // 2))

    @guppy.comptime
    def mod(x: int, y: int) -> int:
        return 15 % (x % (y % 10))

    @guppy.comptime
    def pow(x: int, y: int) -> int:
        return 4 ** (x ** (y**0))

    run_int_fn(pos, 10, [10])
    run_int_fn(neg, -10, [10])
    run_int_fn(add, 2, [3, -4])
    run_int_fn(sub, -8, [3, -4])
    run_int_fn(mul, -24, [3, -4])
    run_int_fn(div, 20, [25, 10])
    run_int_fn(mod, 7, [8, 9])
    run_int_fn(pow, 16, [2, 100])


def test_float(validate, run_float_fn_approx):
    @guppy.comptime
    def pos(x: float) -> float:
        return +x

    @guppy.comptime
    def neg(x: float) -> float:
        return -x

    @guppy.comptime
    def add(x: float, y: float) -> float:
        return 1 + (x + (y + 2))

    @guppy.comptime
    def sub(x: float, y: float) -> float:
        return 1 - (x - (y - 2))

    @guppy.comptime
    def mul(x: float, y: float) -> float:
        return 1 * (x * (y * 2))

    @guppy.comptime
    def div(x: float, y: float) -> float:
        return 100 / (x / (y / 2))

    # TODO: Requires lowering of `ffloor` op: https://github.com/CQCL/hugr/issues/1905
    # @guppy.comptime
    # def floordiv(x: float, y: float) -> float:
    #     return 100 // (x // (y // 2))

    # TODO: Requires lowering of `fpow` op: https://github.com/CQCL/hugr/issues/1905
    # @guppy.comptime
    # def pow(x: float, y: float) -> float:
    #     return 4 ** (x ** (y ** 0.5))

    run_float_fn_approx(pos, 10.5, [10.5])
    run_float_fn_approx(neg, -10.5, [10.5])
    run_float_fn_approx(add, 1.5, [3, -4.5])
    run_float_fn_approx(sub, -8.5, [3, -4.5])
    run_float_fn_approx(mul, -27.0, [3, -4.5])
    run_float_fn_approx(div, 400.0, [0.5, 4])

    # TODO: Requires lowering of `ffloor` op: https://github.com/CQCL/hugr/issues/1905
    # emulate_float_fn_approx(div, ..., [...])

    # TODO: Requires lowering of `fpow` op: https://github.com/CQCL/hugr/issues/1905
    # emulate_float_fn_approx(pow, ..., [...])


def test_angle(validate):
    @guppy.comptime
    def neg(x: angle) -> angle:
        return -x

    @guppy.comptime
    def neg_pi() -> angle:
        return -pi

    @guppy.comptime
    def add(x: angle, y: float) -> angle:
        return pi + (x + angle(y) + pi)

    @guppy.comptime
    def sub(x: float, y: angle) -> angle:
        return pi - (angle(x) - (y - pi))

    @guppy.comptime
    def mul(x: float, y: angle) -> angle:
        return 1.5 * (x * (y * 2))

    @guppy.comptime
    def div(x: angle, y: float) -> angle:
        return x / y

    @guppy
    def main() -> None:
        """Dummy main function"""
        add, sub, mul, div

    validate(main.compile())


def test_dunder_coercions(validate):
    @guppy.comptime
    def test1(x: int) -> float:
        return 1.0 + x

    @guppy.comptime
    def test2(x: int) -> float:
        return x + 1.0

    @guppy.comptime
    def test3(x: float) -> float:
        return 1 + x

    @guppy.comptime
    def test4(x: float) -> float:
        return x + 1

    @guppy.comptime
    def test5(x: int, y: float) -> float:
        return x + y

    @guppy.comptime
    def test6(x: float, y: int) -> float:
        return x + y

    validate(test1.compile())
    validate(test2.compile())
    validate(test3.compile())
    validate(test4.compile())
    validate(test5.compile())
    validate(test6.compile())


def test_const(validate):
    x = guppy.constant("x", "int", IntVal(10, 6))

    @guppy.comptime
    def test1() -> int:
        return -x

    @guppy.comptime
    def test2() -> int:
        return 1 + x

    @guppy.comptime
    def test2() -> int:  # noqa: F811
        return x * 2

    @guppy.comptime
    def test3() -> float:
        return 1.5 - x

    @guppy.comptime
    def test4() -> float:
        return x / 0.5

    validate(test1.compile())
    validate(test2.compile())
    validate(test3.compile())
    validate(test4.compile())



"""Tools for inspecting source code when running in IPython."""


def is_running_ipython() -> bool:
    """Checks if we are currently running in IPython"""
    try:
        return get_ipython() is not None  # type: ignore[name-defined]
    except NameError:
        return False


def normalize_ipython_dummy_files(filename: str) -> str:
    """Turns dummy file names generated for IPython cells into readable names like
    "<In[42]>".

    In vanilla IPython, cells have filenames like "<ipython-input-3-3e9b5833de21>" and
    in Jupyter, cells have filenames like "/var/{...}/ipykernel_82076/61218616.py".
    """
    try:
        if shell := get_ipython():  # type: ignore[name-defined]
            res = shell.compile.format_code_name(filename)
            if res is None:
                return filename
            return f"<{res[1]}>"
        else:
            return filename
    except NameError:
        return filename



from guppylang.decorator import guppy
from guppylang.std.builtins import array
from guppylang.std.quantum import qubit


@guppy.declare
def foo() -> array[qubit, 10]: ...


@guppy
def main() -> qubit:
    return foo()[0]


main.compile()


from typing import no_type_check
from guppylang.decorator import guppy
from guppylang.std.builtins import array, result


@guppy
@no_type_check
def main() -> None:
   xs = array(1, 2, 3)
   ys = xs
   result("foo", xs)


main.compile()



from guppylang import guppy
from guppylang.std.builtins import nat, comptime


@guppy.declare
def foo(n: nat @comptime) -> None: ...


@guppy
def main(b: bool, m: nat) -> None:
    if b:
        foo(m)


main.compile()



import ast
from dataclasses import dataclass, field

from hugr import Node, Wire
from hugr import val as hv
from hugr.build.dfg import DefinitionBuilder, OpVar

from guppylang_internals.ast_util import AstNode
from guppylang_internals.checker.core import Globals
from guppylang_internals.compiler.core import CompilerContext, DFContainer
from guppylang_internals.definition.common import CompilableDef, ParsableDef
from guppylang_internals.definition.value import (
    CompiledHugrNodeDef,
    CompiledValueDef,
    ValueDef,
)
from guppylang_internals.span import SourceMap
from guppylang_internals.tys.parsing import type_from_ast


@dataclass(frozen=True)
class RawConstDef(ParsableDef):
    """A raw constant definition as provided by the user."""

    type_ast: ast.expr
    value: hv.Value

    description: str = field(default="constant", init=False)

    def parse(self, globals: Globals, sources: SourceMap) -> "ConstDef":
        """Parses and checks the user-provided signature of the function."""
        return ConstDef(
            self.id,
            self.name,
            self.defined_at,
            type_from_ast(self.type_ast, globals, {}),
            self.type_ast,
            self.value,
        )


@dataclass(frozen=True)
class ConstDef(RawConstDef, ValueDef, CompilableDef):
    """A constant with a checked type."""

    def compile_outer(
        self, graph: DefinitionBuilder[OpVar], ctx: CompilerContext
    ) -> "CompiledConstDef":
        const_node = graph.add_const(self.value)
        return CompiledConstDef(
            self.id,
            self.name,
            self.defined_at,
            self.ty,
            self.type_ast,
            self.value,
            const_node,
        )


@dataclass(frozen=True)
class CompiledConstDef(ConstDef, CompiledValueDef, CompiledHugrNodeDef):
    """A constant that has been compiled to a Hugr node."""

    const_node: Node

    @property
    def hugr_node(self) -> Node:
        """The Hugr node this definition was compiled into."""
        return self.const_node

    def load(self, dfg: DFContainer, ctx: CompilerContext, node: AstNode) -> Wire:
        """Loads the extern value into a local Hugr dataflow graph."""
        return dfg.builder.load(self.const_node)



from guppylang.decorator import guppy


@guppy.struct
class S:
    x: int

    @guppy
    def foo(self: "S") -> None:
        pass


@guppy.comptime
def test(x: int) -> None:
    s = S(x)
    s.foo = 0


test.compile()



import functools
from collections.abc import Sequence
from typing import cast

from hugr import Wire, ops
from hugr import tys as ht
from hugr.build import cfg as hc
from hugr.build.dfg import DP, DfBase
from hugr.hugr.node_port import ToNode

from guppylang_internals.checker.cfg_checker import (
    CheckedBB,
    CheckedCFG,
    Row,
    Signature,
)
from guppylang_internals.checker.core import Place, Variable
from guppylang_internals.compiler.core import (
    CompilerContext,
    DFContainer,
    is_return_var,
    return_var,
)
from guppylang_internals.compiler.expr_compiler import ExprCompiler
from guppylang_internals.compiler.stmt_compiler import StmtCompiler
from guppylang_internals.std._internal.compiler.tket_bool import OpaqueBool, read_bool
from guppylang_internals.tys.ty import SumType, row_to_type, type_to_row


def compile_cfg(
    cfg: CheckedCFG[Place],
    container: DfBase[DP],
    inputs: Sequence[Wire],
    ctx: CompilerContext,
) -> hc.Cfg:
    """Compiles a CFG to Hugr."""
    # Patch the CFG with dummy return variables
    # TODO: This mutates the CFG in-place which leads to problems when trying to lower
    #  the same function to Hugr twice. For now we just check that the return vars
    #  haven't already been inserted, but we should figure out a better way to handle
    #  this: https://github.com/CQCL/guppylang/issues/428
    if all(
        not is_return_var(v.name)
        for v in cfg.exit_bb.sig.input_row
        if isinstance(v, Variable)
    ):
        insert_return_vars(cfg)

    builder = container.add_cfg(*inputs)

    # Explicitly annotate the output types since Hugr can't infer them if the exit is
    # unreachable
    out_tys = [place.ty.to_hugr(ctx) for place in cfg.exit_bb.sig.input_row]
    # TODO: Use proper API for this once it's added in hugr-py:
    #  https://github.com/CQCL/hugr/issues/1816
    builder._exit_op._cfg_outputs = out_tys
    builder.parent_op._outputs = out_tys
    builder.parent_node = builder.hugr._update_node_outs(
        builder.parent_node, len(out_tys)
    )

    blocks: dict[CheckedBB[Place], ToNode] = {}
    for bb in cfg.bbs:
        blocks[bb] = compile_bb(bb, builder, bb == cfg.entry_bb, ctx)
    for bb in cfg.bbs:
        for i, succ in enumerate(bb.successors):
            builder.branch(blocks[bb][i], blocks[succ])

    return builder


def compile_bb(
    bb: CheckedBB[Place],
    builder: hc.Cfg,
    is_entry: bool,
    ctx: CompilerContext,
) -> ToNode:
    """Compiles a single basic block to Hugr.

    If the basic block is the output block, returns `None`.
    """
    # The exit BB is completely empty
    if bb.is_exit:
        assert len(bb.statements) == 0
        return builder.exit

    # Unreachable BBs (besides the exit) should have been removed by now
    assert bb.reachable

    # Otherwise, we use a regular `Block` node
    block: hc.Block
    inputs: Sequence[Place]
    if is_entry:
        inputs = bb.sig.input_row
        block = builder.add_entry()
    else:
        inputs = sort_vars(bb.sig.input_row)
        block = builder.add_block(*(v.ty.to_hugr(ctx) for v in inputs))

    # Add input node and compile the statements
    dfg = DFContainer(block, ctx)
    for v, wire in zip(inputs, block.input_node, strict=True):
        dfg[v] = wire
    dfg = StmtCompiler(ctx).compile_stmts(bb.statements, dfg)

    # If we branch, we also have to compile the branch predicate
    if len(bb.successors) > 1:
        assert bb.branch_pred is not None
        branch_port = ExprCompiler(ctx).compile(bb.branch_pred, dfg)
        # Convert the bool predicate into a sum for branching.
        pred_ty = builder.hugr.port_type(branch_port.out_port())
        assert pred_ty == OpaqueBool
        branch_port = dfg.builder.add_op(read_bool(), branch_port)
        branch_port = cast(Wire, branch_port)
    else:
        # Even if we don't branch, we still have to add a `Sum(())` predicates
        branch_port = dfg.builder.add_op(ops.Tag(0, ht.UnitSum(1)))

    # Finally, we have to add the block output.
    outputs: Sequence[Place]
    if len(bb.successors) == 1:
        # The easy case is if we don't branch: We just output all variables that are
        # specified by the signature
        [outputs] = bb.sig.output_rows
    else:
        # CFG building ensures that branching BBs don't branch to the exit (exit jumps
        # must always be unconditional)
        assert not any(succ.is_exit for succ in bb.successors)

        # If we branch and the branches use the same places, then we can use a
        # regular output
        first, *rest = bb.sig.output_rows
        if all({p.id for p in first} == {p.id for p in r} for r in rest):
            outputs = first
        else:
            # Otherwise, we have to output a TupleSum: We put all non-linear variables
            # into the branch TupleSum and all linear variables in the normal output
            # (since they are shared between all successors). This is in line with the
            # ordering on variables which puts linear variables at the end.
            # We don't need to worry about the order of return vars since this isn't
            # a branch to an exit (see assert above).
            branch_port = choose_vars_for_tuple_sum(
                unit_sum=branch_port,
                output_vars=[
                    [v for v in sort_vars(row) if v.ty.droppable]
                    for row in bb.sig.output_rows
                ],
                dfg=dfg,
            )
            outputs = [v for v in first if not v.ty.droppable]

    # If this is *not* a jump to the exit BB, we need to sort the outputs to make the
    # signature consistent with what the next BB expects
    if not any(succ.is_exit for succ in bb.successors):
        outputs = sort_vars(outputs)
    else:
        # Exit variables are not allowed to be sorted since their order corresponds to
        # the function outputs
        assert len(bb.successors) == 1, "Exit jumps are always unconditional"

    block.set_block_outputs(branch_port, *(dfg[v] for v in outputs))
    return block


def insert_return_vars(cfg: CheckedCFG[Place]) -> None:
    """Patches a CFG by annotating dummy return variables in the BB signatures.

    The statement compiler turns `return` statements into assignments of dummy variables
    `%ret0`, `%ret1`, etc. We update the exit BB signature to make sure they are
    correctly outputted.
    """
    return_vars = [
        Variable(return_var(i), ty, None)
        for i, ty in enumerate(type_to_row(cfg.output_ty))
    ]
    # Prepend return variables to the exit signature
    cfg.exit_bb.sig = Signature(
        [*return_vars, *cfg.exit_bb.sig.input_row], cfg.exit_bb.sig.output_rows
    )
    # Also patch the predecessors
    for pred in cfg.exit_bb.predecessors:
        # The exit BB will be the only successor
        assert len(pred.sig.output_rows) == 1
        [out_row] = pred.sig.output_rows
        pred.sig = Signature(pred.sig.input_row, [[*return_vars, *out_row]])


def choose_vars_for_tuple_sum(
    unit_sum: Wire, output_vars: list[Row[Place]], dfg: DFContainer
) -> Wire:
    """Selects an output based on a TupleSum.

    Given `unit_sum: Sum(*(), *(), ...)` and output variable rows `#s1, #s2, ...`,
    constructs a TupleSum value of type `Sum(#s1, #s2, ...)`.
    """
    assert all(v.ty.droppable for var_row in output_vars for v in var_row)
    tys = [[v.ty for v in var_row] for var_row in output_vars]
    sum_type = SumType([row_to_type(row) for row in tys]).to_hugr(dfg.ctx)

    # We pass all values into the conditional instead of relying on non-local edges.
    # This is because we can't handle them in lower parts of the stack yet :/
    # TODO: Reinstate use of non-local edges.
    #  See https://github.com/CQCL/guppylang/issues/963
    all_vars = {v.id: dfg[v] for var_row in output_vars for v in var_row}
    all_vars_wires = list(all_vars.values())
    all_vars_idxs = {x: i for i, x in enumerate(all_vars.keys())}

    with dfg.builder.add_conditional(unit_sum, *all_vars_wires) as conditional:
        for i, var_row in enumerate(output_vars):
            with conditional.add_case(i) as case:
                case_inputs = case.inputs()
                outputs = [case_inputs[all_vars_idxs[v.id]] for v in var_row]
                tag = case.add_op(ops.Tag(i, sum_type), *outputs)
                case.set_outputs(tag)
        return conditional


def compare_var(p1: Place, p2: Place) -> int:
    """Defines a `<` order on variables.

    We use this to determine in which order variables are outputted from basic blocks.
    We need to output linear variables at the end, so we do a lexicographic ordering of
    linearity and name.
    """
    return -1 if (not p1.ty.droppable, str(p1)) < (not p2.ty.droppable, str(p2)) else 1


def sort_vars(row: Row[Place]) -> list[Place]:
    """Sorts a row of variables.

    This determines the order in which they are outputted from a BB.
    """
    return sorted(row, key=functools.cmp_to_key(compare_var))



from typing import no_type_check
from guppylang.std.builtins import array
from guppylang.std.debug import state_result
from tests.util import compile_guppy

@compile_guppy
def main(xs: array[int, 3]) -> None:
   state_result("foo", xs)



from guppylang.decorator import guppy

T = guppy.type_var("T")


@guppy.declare
def foo(x: T, y: T) -> None:
    ...


@guppy
def main(x: bool, y: tuple[bool]) -> None:
    foo(x, y)


main.compile()



"""Utilities for iteration over collections of values."""

# mypy: disable-error-code="empty-body, misc, override, valid-type, no-untyped-def"

from __future__ import annotations

from typing import TYPE_CHECKING, Any, no_type_check

from guppylang_internals.decorator import custom_function, extend_type
from guppylang_internals.definition.custom import NoopCompiler
from guppylang_internals.std._internal.checker import RangeChecker
from guppylang_internals.tys.builtin import sized_iter_type_def

from guppylang import guppy
from guppylang.std.option import Option, nothing, some

if TYPE_CHECKING:
    from guppylang.std.lang import owned


L = guppy.type_var("L", copyable=False, droppable=False)
n = guppy.nat_var("n")


@extend_type(sized_iter_type_def)
class SizedIter:
    """A wrapper around an iterator type `L` promising that the iterator will yield
    exactly `n` values.

    Annotating an iterator with an incorrect size is undefined behaviour.
    """

    def __class_getitem__(cls, item: Any) -> type:
        # Dummy implementation to allow subscripting of the `SizedIter` type in
        # positions that are evaluated by the Python interpreter
        return cls

    @custom_function(NoopCompiler())
    def __new__(iterator: L @ owned) -> SizedIter[L, n]:  # type: ignore[type-arg]
        """Casts an iterator into a `SizedIter`."""

    @custom_function(NoopCompiler())
    def unwrap_iter(self: SizedIter[L, n] @ owned) -> L:
        """Extracts the actual iterator."""

    @custom_function(NoopCompiler())
    def __iter__(self: SizedIter[L, n] @ owned) -> SizedIter[L, n]:  # type: ignore[type-arg]
        """Dummy implementation making sized iterators iterable themselves."""


@guppy.struct
class Range:
    next: int
    stop: int

    @guppy
    def __iter__(self: Range) -> Range:
        return self

    @guppy
    @no_type_check
    def __next__(self: Range) -> Option[tuple[int, Range]]:
        if self.next < self.stop:
            return some((self.next, Range(self.next + 1, self.stop)))
        return nothing()


@custom_function(checker=RangeChecker(), higher_order_value=False)
def range(stop: int) -> Range:
    """Limited version of python range().
    Only a single argument (stop/limit) is supported."""



from typing import Generic

from guppylang.decorator import guppy

T = guppy.type_var("T")


@guppy.struct
class MyStruct(Generic[T]):
    x: list[T]


@guppy
def foo(s: MyStruct) -> None:
    pass


foo.compile()



from guppylang import guppy
from guppylang.std.builtins import comptime

T = guppy.type_var("T", copyable=True, droppable=True)


@guppy
def foo(q: T @comptime) -> T:
    return T


@guppy
def main() -> int:
    return foo(42)


main.compile()



from guppylang import guppy
from guppylang.std.builtins import comptime, nat


@guppy
def foo(n: nat @comptime) -> None:
    pass


@guppy
def main(n: nat @ comptime) -> None:
    foo[42](n)


main.compile()



from guppylang.decorator import guppy
from guppylang.std.builtins import owned
from guppylang.std.quantum import qubit


@guppy.struct
class MyStruct:
    q: qubit
    x: int


@guppy
def foo(s: MyStruct @owned) -> int:
    return s.x


foo.compile()



from typing import Callable

from guppylang.decorator import guppy
from guppylang.std.quantum_functional import h

T = guppy.type_var("T")


@guppy.declare
def foo(x: Callable[[T], T]) -> None:
    ...


@guppy
def main() -> None:
    foo(h)


main.compile()



from guppylang import guppy
from guppylang.std.builtins import nat, comptime, array


@guppy.declare
def foo(n: nat @comptime, xs: "array[int, n]") -> None: ...


@guppy
def main(n: nat @comptime, m: nat @comptime) -> None:
    foo(n, array(i for i in range(m)))


main.compile()



from guppylang import qubit, array
from guppylang.decorator import guppy
from guppylang_internals.decorator import custom_function, hugr_op
from guppylang_internals.definition.custom import NoopCompiler
from guppylang_internals.std._internal.util import int_op


def test_basic_overload(validate):
    @guppy.declare
    def variant1(x: float) -> None: ...

    @guppy.declare
    def variant2(x: int, y: int) -> None: ...

    @guppy.declare
    def variant3(x: int, y: int, z: int) -> None: ...

    @guppy.overload(variant1, variant2, variant3)
    def combined(*args): ...

    @guppy
    def main() -> None:
        combined(4.2)
        combined(1, 2)
        combined(1, 2, 3)

    validate(main.compile())


def test_pick_first(validate):
    """Check that we pick the first matching implementation, even if there are other
    matching ones coming afterwards.
    """

    @guppy.declare
    def variant1(x: float) -> int: ...

    @guppy.declare
    def variant2(x: int) -> None: ...

    @guppy.overload(variant1, variant2)
    def combined(*args): ...

    @guppy
    def main() -> int:
        x = combined(42)
        return x

    validate(main.compile())


def test_generic_overload(validate):
    S = guppy.type_var("S")
    T = guppy.type_var("T")

    @guppy.declare
    def variant1(x: T, y: T) -> float: ...

    @guppy.declare
    def variant2(x: T, y: S) -> int: ...

    @guppy.overload(variant1, variant2)
    def combined(*args): ...

    @guppy
    def main() -> int:
        # Check that we pick `variant2` since `variant1` doesn't match (the two args
        # would need to have the same type)
        x = combined(1, 1.0)
        return x

    validate(main.compile())


def test_pick_by_return_type1(validate):
    @guppy.declare
    def variant1() -> int: ...

    @guppy.declare
    def variant2() -> None: ...

    @guppy.overload(variant1, variant2)
    def combined(*args): ...

    @guppy
    def main() -> None:
        out: int = combined()

    validate(main.compile())


def test_pick_by_return_type2(validate):
    T = guppy.type_var("T")

    @guppy.declare
    def variant1(x: T) -> T: ...

    @guppy.declare
    def variant2(x: float) -> int: ...

    @guppy.overload(variant1, variant2)
    def combined(*args): ...

    @guppy
    def main() -> None:
        # Test that we're smart enough to pick variant2. A naive implementation would
        # match `variant1: float -> float` but then fail since it doesn't satisfy the
        # `out: float` annotation. Instead, we consider the whole signature including
        # the return type and infer that `variant2` is the only matching one.
        out: int = combined(42.0)

    validate(main.compile())


def test_comptime_overload_call(validate):
    """Check that overloaded calls work in comptime functions"""

    @guppy.declare
    def variant1(x: float) -> None: ...

    @guppy.declare
    def variant2(x: int, y: int) -> None: ...

    @guppy.declare
    def variant3(x: int, y: int, z: int) -> None: ...

    @guppy.overload(variant1, variant2, variant3)
    def combined(*args): ...

    @guppy.comptime
    def main() -> None:
        combined(4.2)
        combined(1, 2)
        combined(1, 2, 3)

    validate(main.compile())


def test_everything_can_be_overloaded(validate):
    """Test that all kinds of functions can be overloaded."""

    @custom_function(NoopCompiler())
    def custom(a: int) -> int: ...

    @hugr_op(int_op("iadd"))
    def my_hugr_op(a: int, b: int) -> int: ...

    @guppy.declare
    def declaration(a: int, b: int, c: int) -> None: ...

    @guppy
    def defined(a: int, b: int, c: int, d: int) -> int:
        return a + b + c + d

    @guppy.overload(declaration, defined)
    def overloaded(): ...  # Overloaded functions can be overloaded themselves!

    @guppy.overload(custom, my_hugr_op, declaration, defined, overloaded)
    def combined(): ...

    @guppy
    def main() -> None:
        combined(1)
        combined(1, 2)
        combined(1, 2, 3)
        combined(1, 2, 3, 4)

    validate(main.compile())

    # If we have tket installed, we can even overload circuits
    try:
        from pytket import Circuit

        circ = Circuit(1)
        circ.H(0)

        @guppy.pytket(circ)
        def circ1(q: qubit) -> None: ...

        circ2 = guppy.load_pytket("circ2", circ, use_arrays=True)

        @guppy.overload(
            custom, my_hugr_op, declaration, defined, overloaded, circ1, circ2
        )
        def combined(): ...

        @guppy
        def main(q: qubit, qs: array[qubit, 1]) -> None:
            combined(1)
            combined(1, 2)
            combined(1, 2, 3)
            combined(1, 2, 3, 4)
            combined(q)
            combined(qs)

        validate(main.compile())

    except ImportError:
        pass



from guppylang import guppy
from guppylang.std.builtins import panic, exit, comptime
from tests.util import compile_guppy


def test_basic(validate):
    @compile_guppy
    def main() -> None:
        panic("I panicked!")
        exit("I panicked!", 1)

    validate(main)


def test_discard(validate):
    @compile_guppy
    def main() -> None:
        a = 1 + 2
        panic("I panicked!", False, a)
        exit("I exited!", 2, False, a)

    validate(main)


def test_value(validate):
    @guppy
    def foo() -> int:
        return exit("I exited!", 1)

    @guppy
    def bar() -> tuple[int, float]:
        return panic("I panicked!")

    @guppy
    def baz() -> None:
        return panic("I panicked!")

    validate(foo.compile())
    validate(bar.compile())
    validate(baz.compile())


def test_py_message(validate):
    @compile_guppy
    def main(x: int) -> None:
        panic(comptime("I" + "panicked" + "!"))
        exit(comptime("I" + "exited" + "!"), 0)

    validate(main)


def test_comptime_panic(validate):
    @guppy.comptime
    def main() -> None:
        panic("foo")

    validate(main.compile())


def test_comptime_exit(validate):
    @guppy.comptime
    def main() -> None:
        exit("foo", 1)

    validate(main.compile())



from guppylang import qubit
from guppylang.decorator import guppy
from guppylang.std.builtins import owned
from guppylang.std.quantum import discard


@guppy.struct
class S:
    x: int
    q: qubit


@guppy.declare
def bar(s: S) -> None: ...


@guppy.comptime
def foo(s: S @ owned) -> None:
    bar(s)
    # Remains immutable after an inout call
    s.x = 1
    discard(s.q)


foo.compile()



from guppylang.decorator import guppy
from guppylang.std.builtins import owned
from guppylang.std.quantum import qubit


@guppy.declare
def new_qubit() -> qubit:
    ...


@guppy.declare
def measure(q: qubit @owned) -> bool:
    ...


@guppy
def foo(b: bool) -> bool:
    q = new_qubit()
    if b:
        return measure(q)
    return False


foo.compile()



from hugr import tys
from guppylang import guppy
from guppylang_internals.decorator import custom_type


@custom_type(tys.Tuple(), droppable=False)
class NonDroppable:
    pass


@guppy.struct
class MyStruct[D: NonDroppable]:
    pass


MyStruct.compile()



from guppylang.decorator import guppy
from guppylang.std.builtins import owned
from guppylang.std.quantum import qubit


@guppy.declare
def new_qubit() -> qubit:
    ...


@guppy.declare
def measure(q: qubit @owned) -> bool:
    ...


@guppy
def foo(i: int) -> bool:
    b = False
    while i > 0:
        q = new_qubit()
        if i % 10 == 0:
            break
        i -= 1
        b &= measure(q)
    return b


foo.compile()



from guppylang import guppy
from guppylang_internals.decorator import wasm, wasm_module

@wasm_module("", 0)
class Foo:
    @wasm
    def foo(x: int) -> None: ...

@guppy
def main() -> None:
    mod = Foo(0)
    mod.foo(42)
    mod.discard()
    return

main.compile()



from guppylang.decorator import guppy
from guppylang.std.quantum import qubit
from guppylang.std.builtins import owned


@guppy.declare
def bar(q: qubit) -> bool:
    ...


@guppy
def foo(qs: list[tuple[bool, qubit]] @owned) -> list[int]:
    return [42 for b, q in qs if b if bar(q)]


foo.compile()



from guppylang.decorator import guppy
from guppylang.std.quantum import qubit
from guppylang.std.builtins import owned


@guppy
def foo(qs: list[tuple[qubit, qubit]] @owned) -> list[qubit]:
    return [q for q, q in qs]


foo.compile()



from hugr import Wire, ops
from hugr import tys as ht

from guppylang_internals.definition.custom import CustomInoutCallCompiler
from guppylang_internals.definition.value import CallReturnWires
from guppylang_internals.error import InternalGuppyError
from guppylang_internals.nodes import GlobalCall
from guppylang_internals.std._internal.compiler.arithmetic import convert_itousize
from guppylang_internals.std._internal.compiler.prelude import build_unwrap
from guppylang_internals.std._internal.compiler.tket_exts import (
    FUTURES_EXTENSION,
    WASM_EXTENSION,
    ConstWasmModule,
)
from guppylang_internals.tys.builtin import (
    wasm_module_info,
)
from guppylang_internals.tys.ty import (
    FunctionType,
)


class WasmModuleInitCompiler(CustomInoutCallCompiler):
    """Compiler for initialising WASM modules.
    Calls tket's "get_context" and unwraps the `Option` result.
    Returns a `tket.wasm.context` wire.
    """

    def compile_with_inouts(self, args: list[Wire]) -> CallReturnWires:
        # Make a ConstWasmModule as a CustomConst
        assert len(args) == 1
        ctx_arg = args[0]
        ctx_wire = self.builder.add_op(convert_itousize(), ctx_arg)

        ctx_ty = WASM_EXTENSION.get_type("context").instantiate([])
        get_ctx_op = ops.ExtOp(
            WASM_EXTENSION.get_op("get_context"),
            ht.FunctionType([ht.USize()], [ht.Option(ctx_ty)]),
        )
        node = self.builder.add_op(get_ctx_op, ctx_wire)
        opt_w: Wire = node[0]
        err = "Failed to spawn WASM context"
        out_node = build_unwrap(self.builder, opt_w, err)
        return CallReturnWires(regular_returns=[out_node], inout_returns=[])


class WasmModuleDiscardCompiler(CustomInoutCallCompiler):
    """Compiler for discarding WASM contexts."""

    def compile_with_inouts(self, args: list[Wire]) -> CallReturnWires:
        assert len(args) == 1
        ctx = args[0]
        op = WASM_EXTENSION.get_op("dispose_context").instantiate([])
        self.builder.add_op(op, ctx)
        return CallReturnWires(regular_returns=[], inout_returns=[])


class WasmModuleCallCompiler(CustomInoutCallCompiler):
    """Compiler for WASM calls
    When a wasm method is called in guppy, we turn it into 2 tket ops:
    * lookup: wasm.module -> wasm.func
    * call: wasm.context * wasm.func * inputs -> wasm.context * output

    For the wasm.module that we use in lookup, a constant is created for each
    call, using the wasm file information embedded in method's `self` argument.
    """

    fn_name: str

    def __init__(self, name: str) -> None:
        self.fn_name = name

    def compile_with_inouts(self, args: list[Wire]) -> CallReturnWires:
        # The arguments should be:
        # - a WASM context
        # - any args meant for the WASM function
        assert len(args) >= 1
        module_ty = WASM_EXTENSION.get_type("module").instantiate([])
        ctx_ty = WASM_EXTENSION.get_type("context").instantiate([])

        fn_name_arg = ht.StringArg(self.fn_name)
        # Function type without Inout context arg (for building)
        assert isinstance(self.node, GlobalCall)
        assert self.func is not None
        wasm_sig = FunctionType(
            self.func.ty.inputs[1:],
            self.func.ty.output,
        ).to_hugr(self.ctx)

        inputs_row_arg = ht.ListArg([ty.type_arg() for ty in wasm_sig.input])
        output_row_arg = ht.ListArg([ty.type_arg() for ty in wasm_sig.output])

        func_ty = WASM_EXTENSION.get_type("func").instantiate(
            [inputs_row_arg, output_row_arg]
        )
        future_ty = FUTURES_EXTENSION.get_type("Future").instantiate(
            [ht.Tuple(*wasm_sig.output).type_arg()]
        )

        # Get the WASM module information from the type
        selfarg = self.func.ty.inputs[0].ty
        if info := wasm_module_info(selfarg):
            const_module = self.builder.add_const(ConstWasmModule(*info))
        else:
            raise InternalGuppyError(
                "Expected cached signature to have WASM module as first arg"
            )

        wasm_module = self.builder.load(const_module)

        # Lookup the function we want
        wasm_opdef = WASM_EXTENSION.get_op("lookup").instantiate(
            [fn_name_arg, inputs_row_arg, output_row_arg],
            ht.FunctionType([module_ty], [func_ty]),
        )
        wasm_func = self.builder.add_op(wasm_opdef, wasm_module)

        # Call the function
        call_op = WASM_EXTENSION.get_op("call").instantiate(
            [inputs_row_arg, output_row_arg],
            ht.FunctionType([ctx_ty, func_ty, *wasm_sig.input], [ctx_ty, future_ty]),
        )

        ctx, future = self.builder.add_op(call_op, args[0], wasm_func, *args[1:])

        read_opdef = FUTURES_EXTENSION.get_op("Read").instantiate(
            [ht.Tuple(*wasm_sig.output).type_arg()],
            ht.FunctionType([future_ty], [ht.Tuple(*wasm_sig.output)]),
        )
        result = self.builder.add_op(read_opdef, future)
        ws: list[Wire] = list(result[:])
        node = self.builder.add_op(ops.UnpackTuple(wasm_sig.output), *ws)
        ws: list[Wire] = list(node[:])

        return CallReturnWires(regular_returns=ws, inout_returns=[ctx])



from guppylang import guppy
from guppylang_internals.decorator import wasm, wasm_module
from guppylang.std.builtins import nat, comptime
from guppylang.std.qsystem.wasm import spawn_wasm_contexts


def test_wasm_functions(validate):
    @wasm_module("", 42)
    class MyWasm:
        @wasm
        def add_one(self: "MyWasm", x: int) -> int: ...

        @wasm
        def swap(self: "MyWasm", x: int, y: float) -> tuple[float, int]: ...

    @guppy
    def main() -> int:
        [mod1, mod2] = spawn_wasm_contexts(2, MyWasm)
        two = mod1.add_one(1)
        _, two2 = mod2.swap(two, 3.0)
        mod1.discard()
        mod2.discard()
        return two + two2
        return 2

    mod = main.compile()
    validate(mod)


def test_wasm_methods(validate):
    @wasm_module("", 2)
    class MyWasm:
        @wasm
        def foo(self: "MyWasm") -> int: ...

        @guppy
        def bar(self: "MyWasm", x: int) -> int:
            return x + 1

    @guppy
    def main() -> int:
        mod = MyWasm(0)
        x = mod.foo()
        y = mod.bar(x)
        mod.discard()
        return x

    mod = main.compile()
    validate(mod)


def test_wasm_types(validate):
    n = guppy.nat_var("n")

    @wasm_module("", 3)
    class MyWasm:
        @wasm
        def foo(self: "MyWasm", x: tuple[int, tuple[nat, float]], y: int) -> None: ...

    @guppy
    def main() -> None:
        mod = MyWasm(0)
        mod.foo((0, (1, 2.0)), 3)
        mod.discard()
        return

    mod = main.compile()
    validate(mod)


def test_wasm_guppy_module(validate):
    @wasm_module("", 42)
    class MyWasm:
        @wasm
        def add_one(self: "MyWasm", x: int) -> int: ...

        @wasm
        def swap(self: "MyWasm", x: int, y: float) -> tuple[float, int]: ...

    @guppy
    def main() -> int:
        [mod1, mod2] = spawn_wasm_contexts(2, MyWasm)
        two = mod1.add_one(1)
        _, two2 = mod2.swap(two, 3.0)
        mod1.discard()
        mod2.discard()
        return two + two2

    mod = main.compile()
    validate(mod)


def test_comptime(validate):
    @wasm_module("", 42)
    class Foo:
        @wasm
        def goo(self: "Foo") -> int: ...

    def hoo(x: int) -> int:
        return x * x

    @guppy.comptime
    def ioo(x: int) -> int:
        y = hoo(x)
        foo = Foo(0)
        z = foo.goo(y)
        return hoo(z)

    @guppy
    def main() -> int:
        comptime(ioo(42))



from typing import no_type_check
from guppylang.std.builtins import array
from guppylang.std.debug import state_result
from guppylang.std.quantum import qubit
from tests.util import compile_guppy

@compile_guppy
@no_type_check
def main(qs: array[qubit, 3], rs: array[qubit, 4]) -> None:
   state_result("foo", qs, rs)



from guppylang.decorator import guppy

S = guppy.type_var("S")
T = guppy.type_var("T")


@guppy.declare
def foo(x: S, y: T) -> None:
    ...


@guppy
def main(x: int, y: float) -> None:
    foo[float, int](x, y)


main.compile()



from guppylang.decorator import guppy
from guppylang.std.builtins import owned
from guppylang.std.quantum import qubit, measure


@guppy.struct
class MyStruct:
    q1: qubit
    q2: qubit


@guppy
def foo(b: bool, s: MyStruct @owned) -> MyStruct:
    if b:
        measure(s.q2)
    return s


foo.compile()



from guppylang import qubit
from guppylang.decorator import guppy
from guppylang.std.angles import angle
from guppylang.std.builtins import array
from guppylang.std.quantum import h, measure, cx, rz
import itertools


def test_basics(validate):
    @guppy.comptime
    def foo() -> qubit:
        q = qubit()
        h(q)
        bar(q)
        return q

    @guppy.comptime
    def bar(q1: qubit) -> None:
        q2 = qubit()
        cx(q1, q2)
        measure(q2)

    validate(foo.compile())


def test_ladder(validate):
    @guppy.comptime
    def test(qs: array[qubit, 10]) -> None:
        for q1, q2 in itertools.pairwise(qs):
            cx(q1, q2)

    validate(test.compile())


def test_angles(validate):
    @guppy.comptime
    def test(qs: array[qubit, 10], theta: angle) -> None:
        for q in qs:
            rz(q, theta)
            theta /= 2

    validate(test.compile())



from typing import Callable

from guppylang.decorator import guppy
from guppylang.std.builtins import owned
from guppylang.std.quantum import qubit


@guppy.declare
def foo(f: "Callable[[], qubit @owned]") -> None: ...


foo.compile()



from guppylang.decorator import guppy

T = guppy.type_var("T")


@guppy.declare
def foo() -> T:
    ...


@guppy.declare
def bar(x: T, y: T) -> None:
    ...


@guppy
def main() -> None:
    bar(foo(), 42)


main.compile()



import functools
import inspect
import sys
from collections.abc import Callable
from types import FrameType, ModuleType, TracebackType
from typing import ParamSpec, TypeVar

from guppylang_internals.error import GuppyComptimeError, GuppyError, exception_hook

P = ParamSpec("P")
T = TypeVar("T")


def capture_guppy_errors(f: Callable[P, T]) -> Callable[P, T]:
    """Context manager that captures Guppy errors and turns them into runtime
    `GuppyComptimeException`s."""

    @functools.wraps(f)
    def wrapped(*args: P.args, **kwargs: P.kwargs) -> T:
        try:
            return f(*args, **kwargs)
        except GuppyError as err:
            diagnostic = err.error
            msg = diagnostic.rendered_title
            if diagnostic.rendered_span_label:
                msg += f": {diagnostic.rendered_span_label}"
            if diagnostic.rendered_message:
                msg += f"\n{diagnostic.rendered_message}"
            raise GuppyComptimeError(msg) from None

    return wrapped


def hide_trace(f: Callable[P, T]) -> Callable[P, T]:
    """Function decorator that hides compiler-internal frames from the traceback of any
    exception thrown by the decorated function."""

    @functools.wraps(f)
    def wrapped(*args: P.args, **kwargs: P.kwargs) -> T:
        with exception_hook(tracing_except_hook):
            return f(*args, **kwargs)

    return wrapped


def tracing_except_hook(
    excty: type[BaseException], err: BaseException, traceback: TracebackType | None
) -> None:
    """Except hook that removes all compiler-internal frames from the traceback."""
    traceback = remove_internal_frames(traceback)
    try:
        # Check if we're inside a jupyter notebook since it uses its own exception
        # hook. If we're in a regular interpreter, this line will raise a `NameError`
        ipython_shell = get_ipython()  # type: ignore[name-defined]
        ipython_shell.excepthook(excty, err, traceback)
    except NameError:
        sys.__excepthook__(excty, err, traceback)


def get_calling_frame() -> FrameType | None:
    """Finds the first frame that called this function outside the compiler."""
    frame = inspect.currentframe()
    while frame:
        module = inspect.getmodule(frame)
        if module is None or not is_compiler_module(module):
            return frame
        frame = frame.f_back
    return None


def remove_internal_frames(tb: TracebackType | None) -> TracebackType | None:
    """Removes internal frames relating to the Guppy compiler from a traceback."""
    if tb:
        module = inspect.getmodule(tb.tb_frame)
        if module is not None and is_compiler_module(module):
            return remove_internal_frames(tb.tb_next)
        if tb.tb_next:
            tb.tb_next = remove_internal_frames(tb.tb_next)
    return tb


def is_compiler_module(module: ModuleType) -> bool:
    """Checks whether a given Python module belongs to the Guppy compiler."""
    return module.__name__.startswith(
        "guppylang_internals."
    ) or module.__name__.startswith("guppylang.")



from guppylang.decorator import guppy

def test_none(validate, run_int_fn):
    @guppy
    def foo() -> None:
        return None

    @guppy
    def bar(x: None, y: int) -> int:
        return y

    @guppy
    def main() -> int:
        return bar(foo(), 42)

    validate(main.compile())
    run_int_fn(main, expected=42)



from guppylang import qubit
from guppylang.decorator import guppy
from guppylang.std.qsystem import measure


@guppy.comptime
def test() -> int:
    q = qubit()
    if measure(q):
        return 1
    return 0


test.compile()



from abc import ABC, abstractmethod
from collections.abc import Sequence
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, TypeAlias

from hugr import tys as ht
from typing_extensions import Self

from guppylang_internals.ast_util import AstNode
from guppylang_internals.checker.errors.generic import ExpectedError
from guppylang_internals.checker.errors.type_errors import TypeMismatchError
from guppylang_internals.error import GuppyError, GuppyTypeError, InternalGuppyError
from guppylang_internals.tys.arg import Argument, ConstArg, TypeArg
from guppylang_internals.tys.common import ToHugr, ToHugrContext
from guppylang_internals.tys.const import BoundConstVar, ExistentialConstVar
from guppylang_internals.tys.errors import WrongNumberOfTypeArgsError
from guppylang_internals.tys.var import ExistentialVar

if TYPE_CHECKING:
    from guppylang_internals.tys.ty import Type


# We define the `Parameter` type as a union of all `ParameterBase` subclasses defined
# below. This models an algebraic data type and enables exhaustiveness checking in
# pattern matches etc.
# Note that this might become obsolete in case the `@sealed` decorator is added:
#  * https://peps.python.org/pep-0622/#sealed-classes-as-algebraic-data-types
#  * https://github.com/johnthagen/sealed-typing-pep
Parameter: TypeAlias = "TypeParam | ConstParam"


@dataclass(frozen=True)
class ParameterBase(ToHugr[ht.TypeParam], ABC):
    """Abstract base class for parameters used in function and type definitions.

    For example, when defining a struct type

    ```
    @guppy.struct
    class Foo[T, n: int]:
        ...
    ```

    we generate an `StructDef` that depends on the parameters `T` and `n`. From
    this, we obtain a proper `StructType` by providing arguments that are substituted
    for the parameters (for example `Foo[int, 42]`).
    """

    idx: int
    name: str

    @abstractmethod
    def with_idx(self, idx: int) -> Self:
        """Returns a copy of the parameter with a new index."""

    @abstractmethod
    def check_arg(self, arg: Argument, loc: AstNode | None = None) -> Argument:
        """Checks that this parameter can be instantiated with a given argument.

        Raises a user error if the argument is not valid.
        """

    @abstractmethod
    def to_existential(self) -> tuple[Argument, ExistentialVar]:
        """Creates a fresh existential variable that can be instantiated for this
        parameter.

        Returns both the argument and the created variable.
        """

    @abstractmethod
    def to_bound(self, idx: int | None = None) -> Argument:
        """Creates a bound variable with a given index that can be instantiated for this
        parameter.
        """


@dataclass(frozen=True)
class TypeParam(ParameterBase):
    """A parameter of kind type. Used to define generic functions and types."""

    must_be_copyable: bool
    must_be_droppable: bool

    @property
    def can_be_linear(self) -> bool:
        """Whether this type should be treated linearly."""
        return not self.must_be_copyable and not self.must_be_droppable

    def with_idx(self, idx: int) -> "TypeParam":
        """Returns a copy of the parameter with a new index."""
        return TypeParam(idx, self.name, self.must_be_copyable, self.must_be_droppable)

    def check_arg(self, arg: Argument, loc: AstNode | None = None) -> TypeArg:
        """Checks that this parameter can be instantiated with a given argument.

        Raises a user error if the argument is not valid.
        """
        match arg:
            case ConstArg(const):
                err = ExpectedError(loc, "a type", got=f"value of type `{const.ty}`")
                raise GuppyTypeError(err)
            case TypeArg(ty):
                if self.must_be_copyable and not ty.copyable:
                    err = ExpectedError(
                        loc,
                        "a copyable type",
                        got=f"type `{ty}` which is not implicitly copyable",
                    )
                    raise GuppyTypeError(err)
                if self.must_be_droppable and not ty.droppable:
                    err = ExpectedError(
                        loc,
                        "a droppable type",
                        got=f"type `{ty}` which is not implicitly droppable",
                    )
                    raise GuppyTypeError(err)
                return arg

    def to_existential(self) -> tuple[Argument, ExistentialVar]:
        """Creates a fresh existential variable that can be instantiated for this
        parameter.

        Returns both the argument and the created variable.
        """
        from guppylang_internals.tys.ty import ExistentialTypeVar

        var = ExistentialTypeVar.fresh(
            self.name, self.must_be_copyable, self.must_be_droppable
        )
        return TypeArg(var), var

    def to_bound(self, idx: int | None = None) -> TypeArg:
        """Creates a bound variable with a given index that can be instantiated for this
        parameter.
        """
        from guppylang_internals.tys.ty import BoundTypeVar

        if idx is None:
            idx = self.idx
        return TypeArg(
            BoundTypeVar(self.name, idx, self.must_be_copyable, self.must_be_droppable)
        )

    def to_hugr(self, ctx: ToHugrContext) -> ht.TypeParam:
        """Computes the Hugr representation of the parameter."""
        return ht.TypeTypeParam(
            bound=ht.TypeBound.Linear if self.can_be_linear else ht.TypeBound.Copyable
        )

    def __str__(self) -> str:
        """User-facing string representation of the parameter."""
        return self.name


@dataclass(frozen=True)
class ConstParam(ParameterBase):
    """A parameter of kind constant. Used to define fixed-size arrays etc."""

    ty: "Type"

    #: Marker to annotate if this parameter was implicitly generated by a `@comptime`
    #: annotated argument in a function signature.
    from_comptime_arg: bool = field(default=False, kw_only=True)

    def __post_init__(self) -> None:
        if self.ty.unsolved_vars:
            raise InternalGuppyError(
                "Attempted to create constant param with unsolved type"
            )

    def with_idx(self, idx: int) -> "ConstParam":
        """Returns a copy of the parameter with a new index."""
        return ConstParam(idx, self.name, self.ty)

    def check_arg(self, arg: Argument, loc: AstNode | None = None) -> ConstArg:
        """Checks that this parameter can be instantiated with a given argument.

        Raises a user error if the argument is not valid.
        """
        match arg:
            case ConstArg(const):
                if const.ty != self.ty:
                    raise GuppyTypeError(
                        TypeMismatchError(loc, self.ty, const.ty, kind="argument")
                    )
                return arg
            case TypeArg(ty=ty):
                err = ExpectedError(
                    loc, f"expression of type `{self.ty}`", got=f"type `{ty}`"
                )
                raise GuppyTypeError(err)

    def to_existential(self) -> tuple[Argument, ExistentialVar]:
        """Creates a fresh existential variable that can be instantiated for this
        parameter.

        Returns both the argument and the created variable.
        """
        var = ExistentialConstVar.fresh(self.name, self.ty)
        return ConstArg(var), var

    def to_bound(self, idx: int | None = None) -> ConstArg:
        """Creates a bound variable with a given index that can be instantiated for this
        parameter.
        """
        if idx is None:
            idx = self.idx
        return ConstArg(BoundConstVar(self.ty, self.name, idx))

    def to_hugr(self, ctx: ToHugrContext) -> ht.TypeParam:
        """Computes the Hugr representation of the parameter."""
        from guppylang_internals.tys.ty import NumericType

        match self.ty:
            case NumericType(kind=NumericType.Kind.Nat):
                return ht.BoundedNatParam(upper_bound=None)
            case _:
                raise InternalGuppyError(
                    "Tried to convert non-nat const type parameter to Hugr. This "
                    "should have been monomorphized away."
                )

    def __str__(self) -> str:
        """User-facing string representation of the parameter."""
        return f"{self.name}: {self.ty}"


def check_all_args(
    params: Sequence[Parameter],
    args: Sequence[Argument],
    type_name: str,
    loc: AstNode | None = None,
) -> None:
    """Checks a list of arguments against the given parameters.

    Raises a user error if number of arguments doesn't match or one of the argument is
    invalid.
    """
    exp, act = len(params), len(args)
    if exp != act:
        # TODO: Adjust the error span to only point to the offending arguments (similar
        #  to how we deal with call args in the expression checker). This requires
        #  threading the type arg spans down to this point
        raise GuppyError(WrongNumberOfTypeArgsError(loc, exp, act, type_name))

    # Now check that the kinds match up
    for param, arg in zip(params, args, strict=True):
        # TODO: The error location is bad. We want the location of `arg`, not of the
        #  whole thing.
        param.check_arg(arg, loc)



from pytket import Circuit

from guppylang.decorator import guppy
from guppylang.std.quantum import qubit

circ = Circuit(2)
circ.X(0)
circ.Y(1)


guppy_circ = guppy.load_pytket("guppy_circ", circ, use_arrays=False)

@guppy
def foo(q: qubit, p: qubit) -> None:
    guppy_circ(q, p)


foo.compile()


from guppylang.decorator import guppy
from guppylang.std.quantum import qubit


@guppy.struct
class MyStruct:
    q: qubit

@guppy
def test(q: qubit) -> MyStruct:
    return MyStruct(q)


test.compile()



import functools
import itertools
from collections.abc import Callable, Iterator, Sequence
from contextlib import suppress
from dataclasses import dataclass
from pathlib import Path
from typing import Any, ClassVar, NamedTuple, TypeAlias

from hugr import Wire

import guppylang_internals.checker.expr_checker as expr_checker
from guppylang_internals.checker.errors.generic import UnsupportedError
from guppylang_internals.checker.errors.type_errors import (
    BinaryOperatorNotDefinedError,
    UnaryOperatorNotDefinedError,
)
from guppylang_internals.definition.common import DefId, Definition
from guppylang_internals.definition.ty import TypeDef
from guppylang_internals.definition.value import (
    CallableDef,
    CompiledValueDef,
)
from guppylang_internals.engine import DEF_STORE, ENGINE
from guppylang_internals.error import GuppyComptimeError, GuppyError, GuppyTypeError
from guppylang_internals.ipython_inspect import normalize_ipython_dummy_files
from guppylang_internals.tracing.state import get_tracing_state, tracing_active
from guppylang_internals.tracing.util import (
    capture_guppy_errors,
    get_calling_frame,
    hide_trace,
)
from guppylang_internals.tys.ty import FunctionType, StructType, Type

# Mapping from unary dunder method to display name of the operation
unary_table = dict(expr_checker.unary_table.values())

# Mapping from binary dunder method to reversed method and display name of the operation
binary_table = {
    method: (reverse_method, display_name)
    for method, reverse_method, display_name in expr_checker.binary_table.values()
}

# Mapping from reverse binary dunder method to original method and display name of the
# operation
reverse_binary_table = {
    reverse_method: (method, display_name)
    for method, reverse_method, display_name in expr_checker.binary_table.values()
}

UnaryDunderMethod: TypeAlias = Callable[["DunderMixin"], Any]
BinaryDunderMethod: TypeAlias = Callable[["DunderMixin", Any], Any]


def unary_operation(f: UnaryDunderMethod) -> UnaryDunderMethod:
    """Decorator for methods corresponding to unary operations like `__neg__` etc.

    Emits a user error if the unary operation is not defined for the given type.
    """

    @functools.wraps(f)
    @capture_guppy_errors
    def wrapped(self: "DunderMixin") -> Any:
        from guppylang_internals.tracing.state import get_tracing_state
        from guppylang_internals.tracing.unpacking import guppy_object_from_py

        state = get_tracing_state()
        self = guppy_object_from_py(self, state.dfg.builder, state.node, state.ctx)

        with suppress(Exception):
            return f(self)

        raise GuppyTypeError(
            UnaryOperatorNotDefinedError(state.node, self._ty, unary_table[f.__name__])
        )

    return wrapped


def binary_operation(f: BinaryDunderMethod) -> BinaryDunderMethod:
    """Decorator for methods corresponding to binary operations like `__add__` etc.

    Delegate calls to their reversed versions `__radd__` etc. if the original one
    doesn't type check. Otherwise, emits an error informing the user that the binary
    operation is not defined for those types.
    """

    @functools.wraps(f)
    @capture_guppy_errors
    def wrapped(self: "DunderMixin", other: Any) -> Any:
        from guppylang_internals.tracing.state import get_tracing_state
        from guppylang_internals.tracing.unpacking import guppy_object_from_py

        state = get_tracing_state()
        self = guppy_object_from_py(self, state.dfg.builder, state.node, state.ctx)
        other = guppy_object_from_py(other, state.dfg.builder, state.node, state.ctx)

        # First try the method on `self`
        with suppress(Exception):
            return f(self, other)

        # If that failed, try the reverse method on `other`.
        # NB: We know that `f.__name__` is in one of the tables since we make sure to
        # only put this decorator on the correct dunder methods below
        if f.__name__ in binary_table:
            reverse_method, display_name = binary_table[f.__name__]
            left_ty, right_ty = self._ty, other._ty
        else:
            reverse_method, display_name = reverse_binary_table[f.__name__]
            left_ty, right_ty = other._ty, self._ty
        with suppress(Exception):
            return other.__getattr__(reverse_method)(self)

        raise GuppyTypeError(
            BinaryOperatorNotDefinedError(state.node, left_ty, right_ty, display_name)
        )

    return wrapped


class DunderMixin:
    """Mixin class to allow `GuppyObject`s and `GuppyDefinition`s to be used in
    arithmetic expressions etc. via providing the corresponding dunder methods
    delegating to the objects impls.
    """

    def _get_method(self, name: str) -> Any:
        from guppylang_internals.tracing.state import get_tracing_state
        from guppylang_internals.tracing.unpacking import guppy_object_from_py

        state = get_tracing_state()
        self = guppy_object_from_py(self, state.dfg.builder, state.node, state.ctx)
        return self.__getattr__(name)

    def __abs__(self) -> Any:
        return self._get_method("__abs__")()

    @binary_operation
    def __add__(self, other: Any) -> Any:
        return self._get_method("__add__")(other)

    @binary_operation
    def __and__(self, other: Any) -> Any:
        return self._get_method("__and__")(other)

    def __bool__(self: Any) -> Any:
        return self._get_method("__bool__")()

    def __ceil__(self: Any) -> Any:
        return self._get_method("__ceil__")()

    def __divmod__(self, other: Any) -> Any:
        return self._get_method("__divmod__")(other)

    @binary_operation
    def __eq__(self, other: object) -> Any:
        return self._get_method("__eq__")(other)

    def __float__(self) -> Any:
        return self._get_method("__float__")()

    def __floor__(self) -> Any:
        return self._get_method("__floor__")()

    @binary_operation
    def __floordiv__(self, other: Any) -> Any:
        return self._get_method("__floordiv__")(other)

    @binary_operation
    def __ge__(self, other: Any) -> Any:
        return self._get_method("__ge__")(other)

    @binary_operation
    def __gt__(self, other: Any) -> Any:
        return self._get_method("__gt__")(other)

    def __int__(self) -> Any:
        return self._get_method("__int__")()

    @unary_operation
    def __invert__(self) -> Any:
        return self._get_method("__invert__")()

    @binary_operation
    def __le__(self, other: Any) -> Any:
        return self._get_method("__le__")(other)

    @binary_operation
    def __lshift__(self, other: Any) -> Any:
        return self._get_method("__lshift__")(other)

    @binary_operation
    def __lt__(self, other: Any) -> Any:
        return self._get_method("__lt__")(other)

    @binary_operation
    def __mod__(self, other: Any) -> Any:
        return self._get_method("__mod__")(other)

    @binary_operation
    def __mul__(self, other: Any) -> Any:
        return self._get_method("__mul__")(other)

    @binary_operation
    def __ne__(self, other: object) -> Any:
        return self._get_method("__ne__")(other)

    @unary_operation
    def __neg__(self) -> Any:
        return self._get_method("__neg__")()

    @binary_operation
    def __or__(self, other: Any) -> Any:
        return self._get_method("__or__")(other)

    @unary_operation
    def __pos__(self) -> Any:
        return self._get_method("__pos__")()

    @binary_operation
    def __pow__(self, other: Any) -> Any:
        return self._get_method("__pow__")(other)

    @binary_operation
    def __radd__(self, other: Any) -> Any:
        return self._get_method("__radd__")(other)

    @binary_operation
    def __rand__(self, other: Any) -> Any:
        return self._get_method("__rand__")(other)

    @binary_operation
    def __rfloordiv__(self, other: Any) -> Any:
        return self._get_method("__rfloordiv__")(other)

    @binary_operation
    def __rlshift__(self, other: Any) -> Any:
        return self._get_method("__rlshift__")(other)

    @binary_operation
    def __rmod__(self, other: Any) -> Any:
        return self._get_method("__rmod__")(other)

    @binary_operation
    def __rmul__(self, other: Any) -> Any:
        return self._get_method("__rmul__")(other)

    @binary_operation
    def __ror__(self, other: Any) -> Any:
        return self._get_method("__ror__")(other)

    @binary_operation
    def __rpow__(self, other: Any) -> Any:
        return self._get_method("__rpow__")(other)

    @binary_operation
    def __rrshift__(self, other: Any) -> Any:
        return self._get_method("__pow__")(other)

    @binary_operation
    def __rshift__(self, other: Any) -> Any:
        return self._get_method("__rshift__")(other)

    @binary_operation
    def __rsub__(self, other: Any) -> Any:
        return self._get_method("__rsub__")(other)

    @binary_operation
    def __rtruediv__(self, other: Any) -> Any:
        return self._get_method("__rtruediv__")(other)

    @binary_operation
    def __rxor__(self, other: Any) -> Any:
        return self._get_method("__rxor__")(other)

    @binary_operation
    def __sub__(self, other: Any) -> Any:
        return self._get_method("__sub__")(other)

    @binary_operation
    def __truediv__(self, other: Any) -> Any:
        return self._get_method("__truediv__")(other)

    def __trunc__(self) -> Any:
        return self._get_method("__trunc__")()

    @binary_operation
    def __xor__(self, other: Any) -> Any:
        return self._get_method("__xor__")(other)


class ObjectUse(NamedTuple):
    """Records a use of a non-copyable `GuppyObject`."""

    #: Path of the Python file in which the use occurred
    module: str

    #: Line number of the use
    lineno: int

    #: If the use was as an argument to a Guppy function, we also record a reference to
    #: the called function.
    called_func: CallableDef | None


@dataclass(frozen=True)
class GuppyObjectId:
    """Unique id for abstract GuppyObjects allocated during tracing."""

    id: int

    _fresh_ids: ClassVar[Iterator[int]] = itertools.count()

    @classmethod
    def fresh(cls) -> "GuppyObjectId":
        return GuppyObjectId(next(cls._fresh_ids))


class GuppyObject(DunderMixin):
    """The runtime representation of abstract Guppy objects during tracing.

    They correspond to a single Hugr wire within the current dataflow graph.
    """

    #: The type of this object
    _ty: Type

    #: The Hugr wire holding this object
    _wire: Wire

    #: Whether this object has been used
    _used: ObjectUse | None

    #: Unique id for this object
    _id: GuppyObjectId

    def __init__(self, ty: Type, wire: Wire, used: ObjectUse | None = None) -> None:
        self._ty = ty
        self._wire = wire
        self._used = used
        self._id = GuppyObjectId.fresh()
        state = get_tracing_state()
        if not ty.droppable and not self._used:
            state.unused_undroppable_objs[self._id] = self

    @hide_trace
    def __getattr__(self, key: str) -> Any:  # type: ignore[misc]
        # Guppy objects don't have fields (structs are treated separately below), so the
        # only attributes we have to worry about are methods.
        func = get_tracing_state().globals.get_instance_func(self._ty, key)
        if func is None:
            raise GuppyComptimeError(
                f"Expression of type `{self._ty}` has no attribute `{key}`"
            )
        return lambda *xs: TracingDefMixin(func)(self, *xs)

    @hide_trace
    def __bool__(self) -> Any:
        err = (
            "Can't branch on a dynamic Guppy value since its concrete value is not "
            "known at comptime. Consider defining a regular Guppy function to perform "
            "dynamic branching."
        )
        raise GuppyComptimeError(err)

    @hide_trace
    @capture_guppy_errors
    def __call__(self, *args: Any) -> Any:
        if not isinstance(self._ty, FunctionType):
            err = f"Value of type `{self._ty}` is not callable"
            raise GuppyComptimeError(err)

        # TODO: Support higher-order functions
        state = get_tracing_state()
        raise GuppyError(
            UnsupportedError(state.node, "Higher-order comptime functions")
        )

    @hide_trace
    def __iter__(self) -> Any:
        # Abstract Guppy objects are not iterable from Python since our iterator
        # protocol doesn't work during tracing.
        raise GuppyComptimeError(
            f"Expression of type `{self._ty}` is not iterable at comptime"
        )

    def _use_wire(self, called_func: CallableDef | None) -> Wire:
        # Panic if the value has already been used
        if self._used and not self._ty.copyable:
            use = self._used
            # TODO: Should we print the full path to the file or only the name as is
            #  done here? Note that the former will lead to challenges with golden
            #  tests
            filename = Path(normalize_ipython_dummy_files(use.module)).name
            err = (
                f"Value with non-copyable type `{self._ty}` was already used\n\n"
                f"Previous use occurred in {filename}:{use.lineno}"
            )
            if use.called_func:
                err += f" as an argument to `{use.called_func.name}`"
            raise GuppyComptimeError(err)
        # Otherwise, mark it as used
        else:
            frame = get_calling_frame()
            assert frame is not None
            module_name = frame.f_code.co_filename
            self._used = ObjectUse(module_name, frame.f_lineno, called_func)
            if not self._ty.droppable:
                state = get_tracing_state()
                state.unused_undroppable_objs.pop(self._id)
        return self._wire


class GuppyStructObject(DunderMixin):
    """The runtime representation of Guppy struct objects during tracing.

    Note that `GuppyStructObject` is not a `GuppyObject` itself since it's not backed
    by a single wire, but it can contain multiple of them.

    Mutation of structs during tracing is generally unchecked. We allow users to write
    whatever they want into the fields, making it more or less isomorphic to a Python
    dataclass. Thus, structs need to be checked at function call boundaries to ensure
    that the user hasn't messed up. This is done in `guppylang.tracing.unpacking.
    guppy_object_from_py`.

    Similar to dataclasses, we allow structs to be `frozen` which makes them immutable.
    This is needed to preserve Python semantics when structs are used as non-borrowed
    function arguments: Mutation in the function body cannot be observed from the
    outside, so we prevent it to avoid confusion.
    """

    #: The type of this struct object
    _ty: StructType

    #: Mapping from field names to values. The values can be any Python object.
    _field_values: dict[str, Any]

    #: Whether this struct object is frozen, i.e. immutable
    _frozen: bool

    def __init__(
        self, ty: StructType, field_values: Sequence[Any], frozen: bool
    ) -> None:
        field_values_dict = {
            f.name: v for f, v in zip(ty.fields, field_values, strict=True)
        }
        # Can't use regular assignment for class attributes since we override
        # `__setattr__` below
        object.__setattr__(self, "_ty", ty)
        object.__setattr__(self, "_field_values", field_values_dict)
        object.__setattr__(self, "_frozen", frozen)

    @hide_trace
    def __getattr__(self, key: str) -> Any:  # type: ignore[misc]
        # It could be an attribute
        if key in self._field_values:
            return self._field_values[key]
        # Or a method
        func = get_tracing_state().globals.get_instance_func(self._ty, key)
        if func is None:
            err = f"Expression of type `{self._ty}` has no attribute `{key}`"
            raise AttributeError(err)
        return lambda *xs: TracingDefMixin(func)(self, *xs)

    @hide_trace
    def __setattr__(self, key: str, value: Any) -> None:
        if key in self._field_values:
            if self._frozen:
                err = (
                    f"Object of type `{self._ty}` is an owned function argument. "
                    "Therefore, this mutation won't be visible to the caller."
                )
                raise GuppyComptimeError(err)
            self._field_values[key] = value
        else:
            err = f"Expression of type `{self._ty}` has no attribute `{key}`"
            raise AttributeError(err)

    @hide_trace
    def __iter__(self) -> Any:
        # Abstract Guppy objects are not iterable from Python since our iterator
        # protocol doesn't work during tracing.
        raise GuppyComptimeError(f"Expression of type `{self._ty}` is not iterable")


@dataclass(frozen=True)
class TracingDefMixin(DunderMixin):
    """Mixin to provide tracing semantics for definitions."""

    wrapped: Definition

    @property
    def id(self) -> DefId:
        return self.wrapped.id

    @hide_trace
    def __call__(self, *args: Any) -> Any:
        from guppylang_internals.tracing.function import trace_call

        if not tracing_active():
            raise GuppyComptimeError(
                f"{self.wrapped.description.capitalize()} `{self.wrapped.name}` may "
                "only be called in a Guppy context"
            )

        defn = ENGINE.get_checked(self.wrapped.id)
        if isinstance(defn, CallableDef):
            return trace_call(defn, *args)
        elif (
            isinstance(defn, TypeDef)
            and defn.id in DEF_STORE.impls
            and "__new__" in DEF_STORE.impls[defn.id]
        ):
            constructor = DEF_STORE.raw_defs[DEF_STORE.impls[defn.id]["__new__"]]
            return TracingDefMixin(constructor)(*args)
        err = f"{defn.description.capitalize()} `{defn.name}` is not callable"
        raise GuppyComptimeError(err)

    def __getitem__(self, item: Any) -> Any:
        # If this is a type definition, then `__getitem__` might be called when
        # specifying generic arguments
        if isinstance(self.wrapped, TypeDef):
            # It doesn't really matter what we return here since we don't support types
            # as comptime values yet, so just give back the definition
            return self
        # TODO: Alternatively, it could be a type application on a generic function.
        #  Supporting those requires a comptime representation of types as values
        if tracing_active():
            state = get_tracing_state()
            defn = state.globals[self.wrapped.id]
            if isinstance(defn, CallableDef) and defn.ty.parametrized:
                raise GuppyComptimeError(
                    "Explicitly specifying type arguments of generic functions in a "
                    "comptime context is not supported yet"
                )
        raise GuppyComptimeError(
            f"{self.wrapped.description.capitalize()} `{self.wrapped.name}` is not "
            "subscriptable"
        )

    def to_guppy_object(self) -> GuppyObject:
        state = get_tracing_state()
        defn, [] = state.ctx.build_compiled_def(self.id, type_args=[])
        if isinstance(defn, CompiledValueDef):
            wire = defn.load(state.dfg, state.ctx, state.node)
            return GuppyObject(defn.ty, wire, None)
        elif isinstance(defn, TypeDef):
            if defn.id in DEF_STORE.impls and "__new__" in DEF_STORE.impls[defn.id]:
                constructor = DEF_STORE.raw_defs[DEF_STORE.impls[defn.id]["__new__"]]
                return TracingDefMixin(constructor).to_guppy_object()
        err = f"{defn.description.capitalize()} `{defn.name}` is not a value"
        raise GuppyComptimeError(err)



from guppylang.decorator import guppy
from guppylang.std.builtins import array, owned, mem_swap
from guppylang.std.quantum import qubit, discard, h, x


def test_indexing(run_int_fn):
    @guppy
    def main() -> int:
        t = (0.2, 2, 2.2)
        x = t[1]
        return x

    run_int_fn(main, 2)


def test_indexing_input(validate):
    @guppy
    def main(t: tuple[int, float]) -> int:
        return t[0]

    validate(main.compile())


def test_indexing_drop(run_int_fn):
    @guppy
    def main() -> int:
        return (1, 2, 3)[2]

    run_int_fn(main, 3)


def test_indexing_drop2(validate):
    @guppy
    def main() -> qubit:
        return (1, 2, qubit())[2]

    validate(main.compile())


def test_indexing_nesting(run_int_fn):
    @guppy
    def main() -> int:
        t = (1, 2, (3, 4))
        return t[2][0]

    run_int_fn(main, 3)


def test_indexing_struct(run_int_fn):
    @guppy.struct
    class S:
        t: tuple[array[tuple[int, int], 1], int]

    @guppy
    def main() -> int:
        s = S((array((1, 2)), 3))
        return s.t[0][0][0] + 1

    run_int_fn(main, 2)


def test_rest_after_use(validate):
    @guppy.declare
    def use(x: qubit @ owned) -> None: ...

    @guppy
    def main() -> int:
        t = (1, qubit())
        use(t[1])
        return t[0]

    validate(main.compile())


def test_control_flow(validate):
    @guppy
    def foo(b: bool) -> None:
        t = (qubit(), qubit(), 0)
        if b:
            h(t[0])
        else:
            h(t[1])
        discard(t[0])
        discard(t[1])

    validate(foo.compile())


def test_control_flow2(validate):
    @guppy
    def foo() -> None:
        ts = array((qubit(), qubit()))
        h(ts[0][0])
        h(ts[0][1])
        for q1, q2 in ts:
            discard(q1)
            discard(q2)

    validate(foo.compile())


def test_mem_swap(validate):
    @guppy
    def foo(b: bool) -> None:
        t = (qubit(), qubit(), 0)
        if b:
            q = qubit()
            mem_swap(t[0], q)
            discard(q)
        else:
            q = qubit()
            mem_swap(t[1], q)
            discard(q)
        h(t[0])
        discard(t[0])
        discard(t[1])

    validate(foo.compile())


def test_non_terminating(validate):
    @guppy
    def foo(t: tuple[qubit, qubit, int]) -> None:
        while True:
            x(t[0])

    validate(foo.compile())



from collections.abc import Callable

from guppylang.decorator import guppy


@guppy
def foo(x: int) -> int:
    def bar(f: Callable[[int], bool]) -> bool:
        return f(42)

    return bar(foo)


foo.compile()



from guppylang.decorator import guppy
from guppylang.std.quantum import qubit


@guppy.struct
class MyStruct:
    q: qubit


@guppy
def foo(b: bool) -> int:
    if b:
        s = MyStruct(qubit())
    else:
        s = MyStruct(qubit())
    return 42


foo.compile()



import ast
from dataclasses import dataclass, field
from typing import ClassVar

from hugr import Node, Wire
from hugr.build import function as hf
from hugr.build.dfg import DefinitionBuilder, OpVar

from guppylang_internals.ast_util import AstNode, has_empty_body, with_loc, with_type
from guppylang_internals.checker.core import Context, Globals
from guppylang_internals.checker.expr_checker import check_call, synthesize_call
from guppylang_internals.checker.func_checker import check_signature
from guppylang_internals.compiler.core import (
    CompilerContext,
    DFContainer,
    requires_monomorphization,
)
from guppylang_internals.definition.common import CompilableDef, ParsableDef
from guppylang_internals.definition.function import (
    PyFunc,
    compile_call,
    load_with_args,
    parse_py_func,
)
from guppylang_internals.definition.value import (
    CallableDef,
    CallReturnWires,
    CompiledCallableDef,
    CompiledHugrNodeDef,
)
from guppylang_internals.diagnostic import Error
from guppylang_internals.error import GuppyError
from guppylang_internals.nodes import GlobalCall
from guppylang_internals.span import SourceMap
from guppylang_internals.tys.param import Parameter
from guppylang_internals.tys.subst import Inst, Subst
from guppylang_internals.tys.ty import Type


@dataclass(frozen=True)
class BodyNotEmptyError(Error):
    title: ClassVar[str] = "Unexpected function body"
    span_label: ClassVar[str] = "Body of declared function `{name}` must be empty"
    name: str


@dataclass(frozen=True)
class MonomorphizeError(Error):
    title: ClassVar[str] = "Invalid function declaration"
    span_label: ClassVar[str] = (
        "Function declaration `{name}` is not allowed to be generic over `{param}`"
    )
    name: str
    param: Parameter


@dataclass(frozen=True)
class RawFunctionDecl(ParsableDef):
    """A raw function declaration provided by the user.

    The raw declaration stores exactly what the user has written (i.e. the AST), without
    any additional checking or parsing.
    """

    python_func: PyFunc
    description: str = field(default="function", init=False)

    def parse(self, globals: Globals, sources: SourceMap) -> "CheckedFunctionDecl":
        """Parses and checks the user-provided signature of the function."""
        func_ast, docstring = parse_py_func(self.python_func, sources)
        ty = check_signature(func_ast, globals)
        if not has_empty_body(func_ast):
            raise GuppyError(BodyNotEmptyError(func_ast.body[0], self.name))
        # Make sure we won't need monomorphization to compile this declaration
        for param in ty.params:
            if requires_monomorphization(param):
                raise GuppyError(MonomorphizeError(func_ast, self.name, param))
        return CheckedFunctionDecl(
            self.id,
            self.name,
            func_ast,
            ty,
            self.python_func,
            docstring,
        )


@dataclass(frozen=True)
class CheckedFunctionDecl(RawFunctionDecl, CompilableDef, CallableDef):
    """A function declaration with parsed and checked signature.

    In particular, this means that we have determined a type for the function.
    """

    defined_at: ast.FunctionDef
    docstring: str | None

    def check_call(
        self, args: list[ast.expr], ty: Type, node: AstNode, ctx: Context
    ) -> tuple[ast.expr, Subst]:
        """Checks the return type of a function call against a given type."""
        # Use default implementation from the expression checker
        args, subst, inst = check_call(self.ty, args, ty, node, ctx)
        node = with_loc(node, GlobalCall(def_id=self.id, args=args, type_args=inst))
        return node, subst

    def synthesize_call(
        self, args: list[ast.expr], node: AstNode, ctx: Context
    ) -> tuple[GlobalCall, Type]:
        """Synthesizes the return type of a function call."""
        # Use default implementation from the expression checker
        args, ty, inst = synthesize_call(self.ty, args, node, ctx)
        node = with_loc(node, GlobalCall(def_id=self.id, args=args, type_args=inst))
        return with_type(ty, node), ty

    def compile_outer(
        self, module: DefinitionBuilder[OpVar], ctx: CompilerContext
    ) -> "CompiledFunctionDecl":
        """Adds a Hugr `FuncDecl` node for this function to the Hugr."""
        assert isinstance(
            module, hf.Module
        ), "Functions can only be declared in modules"
        module: hf.Module = module

        node = module.declare_function(self.name, self.ty.to_hugr_poly(ctx))
        return CompiledFunctionDecl(
            self.id,
            self.name,
            self.defined_at,
            self.ty,
            self.python_func,
            self.docstring,
            node,
        )


@dataclass(frozen=True)
class CompiledFunctionDecl(
    CheckedFunctionDecl, CompiledCallableDef, CompiledHugrNodeDef
):
    """A function declaration with a corresponding Hugr node."""

    declaration: Node

    @property
    def hugr_node(self) -> Node:
        """The Hugr node this definition was compiled into."""
        return self.declaration

    def load_with_args(
        self,
        type_args: Inst,
        dfg: DFContainer,
        ctx: CompilerContext,
        node: AstNode,
    ) -> Wire:
        """Loads the function as a value into a local Hugr dataflow graph."""
        # Use implementation from function definition.
        return load_with_args(type_args, dfg, self.ty, self.declaration)

    def compile_call(
        self,
        args: list[Wire],
        type_args: Inst,
        dfg: DFContainer,
        ctx: CompilerContext,
        node: AstNode,
    ) -> CallReturnWires:
        """Compiles a call to the function."""
        # Use implementation from function definition.
        return compile_call(args, type_args, dfg, self.ty, self.declaration)



from guppylang import array
from guppylang.decorator import guppy

T = guppy.type_var("T", copyable=True, droppable=False)


@guppy
def foo(x: T) -> T:
    return x


@guppy
def main() -> None:
    f = foo[array[int, 10]]


main.compile()



from guppylang.decorator import guppy
from guppylang.std.builtins import owned
from guppylang.std.quantum import qubit


@guppy
def foo(q: qubit @owned) -> int:
    x = q
    return 42


foo.compile()
